# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# popen.py - process control using notifier
# -----------------------------------------------------------------------------
# $Id: popen.py 3104 2008-02-15 14:33:33Z tack $
#
# This module defines a process class similar to the once defined in popen2
# except that this class is aware of the notifier loop.
#
# When creating a Process object you can add files for logging the stdout and
# stderr of the process, you can send data to it and add a callback to be
# called when the process is dead. See the class member functions for more
# details.
#
# By inherting from the class you can also override the functions stdout_cb
# and stderr_cb to process stdout and stderr line by line.
#
# The kill_all function of this class can be called at the end of the programm
# to stop all running processes.
#
# -----------------------------------------------------------------------------
# kaa.notifier - Mainloop and callbacks
# Copyright (C) 2005-2008 Dirk Meyer, Jason Tackaberry, et al.
#
# First Version: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#
# Based on code by Krister Lagerstrom and Andreas Büsching
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Process', 'proclist' ]

# python imports
import os
import fcntl
import popen2
import glob
import re
import logging

# notifier imports
import nf_wrapper as notifier
from callback import Callback
from signals import Signal
from thread import MainThreadCallback, is_mainthread
from async import InProgress
from sockets import IOMonitor, IO_WRITE

# get logging object
log = logging.getLogger('notifier')

# FIXME: rewrite :)

class Process(object):
    """
    Base class for started child processes
    """
    def __init__( self, cmd, debugname = None ):
        """
        Init the child process 'cmd'. This can either be a string or a list
        of arguments (similar to popen2). If debugname is given, the stdout
        and stderr will also be written.
        """

        # Setup signal handlers for the process; allows the class to be
        # useful without subclassing.
        self.signals = {
            "stderr": Signal(),
            "stdout": Signal(),
            "raw-stderr": Signal(),
            "raw-stdout": Signal(),
            "completed": Signal(),
        }

        self._cmd = self._normalize_cmd(cmd)
        self._stop_cmd = None
        self._debugname = debugname
        self.__dead = True
        self.stopping = False
        self.__kill_timer = None
        self.child = None
        self.in_progress = None
        self._write_buffer = []
        self._wmon = None
        self._close_stdin = False


    def _normalize_cmd(self, cmd):
        """
        Converts a command string into a list while honoring quoting, or
        removes empty strings if the cmd is a list.
        """
        if cmd == None:
            return []
        if isinstance(cmd, tuple):
            cmd = list(cmd)
        if isinstance(cmd, list):
            # Remove empty strings from argument list.
            while '' in cmd:
                cmd.remove('')
            return cmd

        assert(isinstance(cmd, str))

        # This might be how you'd do it in C. :)
        cmdlist = []
        curarg = ""
        waiting = None
        last = None
        for c in cmd:
            if (c == ' ' and not waiting) or c == waiting:
                if curarg:
                    cmdlist.append(curarg)
                    curarg = ""
                waiting = None
            elif c in ("'", '"') and not waiting and last != '\\':
                waiting = c
            else:
                curarg += c
            last = c

        if curarg:
            cmdlist.append(curarg)

        return cmdlist


    def start(self, args = None):
        """
        Starts the process.  If args is not None, it can be either a list or
        string, as with the constructor, and is appended to the command line
        specified in the constructor.
        """
        if not self.__dead:
            raise SystemError, "Process is already running."
        if self.stopping:
            raise SystemError, "Process isn't done stopping yet."

        cmd = self._cmd + self._normalize_cmd(args)
        self.__kill_timer = None
        self.__dead = False
        self.binary = cmd[0]

        self.child = popen2.Popen3( cmd, True, 100 )

        flags = fcntl.fcntl(self.child.tochild.fileno(), fcntl.F_GETFL)
        fcntl.fcntl( self.child.tochild.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK )
        self._wmon = IOMonitor(self._handle_write)
        if self._write_buffer:
            self._wmon.register(self.child.tochild, IO_WRITE)

        log.info('running %s (pid=%s)' % ( self.binary, self.child.pid ) )

        # IO_Handler for stdout
        self.stdout = IO_Handler( 'stdout', self.child.fromchild,
                                  self.signals["stdout"], self.signals['raw-stdout'], self._debugname )
        # IO_Handler for stderr
        self.stderr = IO_Handler( 'stderr', self.child.childerr,
                                  self.signals["stderr"], self.signals['raw-stderr'], self._debugname )

        # add child to watcher
        if not is_mainthread():
            MainThreadCallback(proclist.append)(self, self.__child_died)
        else:
            proclist.append( self, self.__child_died )
        self.in_progress = InProgress()
        return self.in_progress


    def get_pid(self):
        """
        Returns the pid of the child process if it has been spawned.  Otherwise
        returns None
        """
        if self.child:
            return self.child.pid


    def close_stdin(self):
        """
        Closes stdin either now, or once the write buffer has been flushed.
        This might be necessary when, for example, we know we have nothing
        more to send to the child, but the child is waiting for either more
        data or a closed fd before outputting data we're interested in.
        """
        if not self.child:
            return

        if self._wmon and self._wmon.active():
            self._close_stdin = True
        else:
            self.child.tochild.close()


    def write(self, data):
        """
        Queue data for writing when the child is ready to receive it.
        """
        if self.child.tochild.closed:
            raise ValueError("Can't write when stdin has been closed")

        self._write_buffer.append(data)
        if self.child and self._wmon and not self._wmon.active():
            self._wmon.register(self.child.tochild, IO_WRITE)


    def _handle_write(self):
        if not self.child:
            # The child process might have died before we had a chance to
            # flush the write buffer.
            del self._write_buffer[:]
            return False

        try:
            while self._write_buffer:
                data = self._write_buffer[0]
                sent = os.write(self.child.tochild.fileno(), data)
                # If we got here, no exception was raised, so we can pop the
                # data from the buffer.
                self._write_buffer.pop(0)
                if sent != len(data):
                    # Not all data was sent, so push the remaining bytes back and
                    # abort the loop.
                    self._write_buffer.insert(0, data[sent:])
                    break
        except IOError, (errno, msg):
            if errno == 11:
                # Resource temporarily unavailable -- trying to write too
                # much data.
                return

        if not self._write_buffer:
            if self._close_stdin:
                self.child.tochild.close()
            return False


    def is_alive( self ):
        """
        Return True if the app is still running
        """
        return self.child and not self.__dead


    def set_stop_command(self, cmd):
        """
        Sets the stop command for this process.  The argument 'cmd' can be
        either a callable or a string.  The command is invoked (if cmd is
        a callback) or the command is written to the child's stdin (if cmd
        is a string or unicode) when the process is being terminated with
        a call to stop().

        Shutdown handlers for the process should be set with this method.
        """
        assert(callable(cmd) or type(cmd) in (str, unicode) or cmd == None)
        self._stop_cmd = cmd


    def stop( self, cmd = None ):
        """
        Stop the child. If 'cmd' is given, this stop command will send to
        the app to stop itself. If this is not working, kill -15 and kill -9
        will be used to kill the app.
        """
        if self.stopping:
            return
        if not is_mainthread():
            return MainThreadCallback(self.stop, cmd)()

        self.stopping = True
        cmd = cmd or self._stop_cmd

        if self.is_alive() and not self.__kill_timer:
            if cmd:
                log.info('sending exit command to app')
                if callable(cmd):
                    cmd()
                else:
                    self.write(cmd)

                cb = Callback( self.__kill, 15 )
                self.__kill_timer = notifier.timer_add( 3000, cb )
            else:
                cb = Callback( self.__kill, 15 )
                self.__kill_timer = notifier.timer_add( 0, cb )


    def __kill( self, signal ):
        """
        Internal kill helper function
        """
        if not self.is_alive():
            self.__dead = True
            self.stopping = False
            return False
        # child needs some assistance with dying ...
        try:
            os.kill( self.child.pid, signal )
        except OSError:
            pass

        if signal == 15:
            cb = Callback( self.__kill, 9 )
        else:
            cb = Callback( self.__killall, 15 )

        self.__kill_timer = notifier.timer_add( 3000, cb )
        return False


    def __killall( self, signal ):
        """
        Internal killall helper function
        """
        if not self.is_alive():
            self.__dead = True
            self.stopping = False
            return False
        # child needs some assistance with dying ...
        try:
            # kill all applications with the string <appname> in their
            # commandline. This implementation uses the /proc filesystem,
            # it is Linux-dependent.
            unify_name = re.compile('[^A-Za-z0-9]').sub
            appname = unify_name('', self.binary)

            cmdline_filenames = glob.glob('/proc/[0-9]*/cmdline')

            for cmdline_filename in cmdline_filenames:
                try:
                    fd = open(cmdline_filename)
                    cmdline = fd.read()
                    fd.close()
                except IOError:
                    continue
                if unify_name('', cmdline).find(appname) != -1:
                    # Found one, kill it
                    pid = int(cmdline_filename.split('/')[2])
                    try:
                        os.kill(pid, signal)
                    except (KeyboardInterrupt, SystemExit), e:
                        os.kill(pid, signal)
                        sys.exit(0)
                    except:
                        pass
        except OSError:
            pass

        log.info('kill -%d %s' % ( signal, self.binary ))
        if signal == 15:
            cb = Callback( self.__killall, 9 )
            self.__kill_timer = notifier.timer_add( 2000, cb )
        else:
            log.critical('PANIC %s' % self.binary)

        return False


    def __child_died( self, pid, status ):
        """
        Callback from watcher when the child died.
        """
        if pid != self.get_pid():
            # We received notification from the watcher concerning another
            # process, so ignore.
            return

        self.__dead = True
        self.stopping = False
        # close IO handler and kill timer
        self.stdout.flush()
        self.stderr.flush()
        self.child.tochild.close()
        self.child = None
        if self.__kill_timer:
            notifier.timer_remove( self.__kill_timer )
        self.in_progress.finish(status >> 8)
        self.in_progress = None
        self.signals['completed'].emit(status >> 8)



class IO_Handler(object):
    """
    Reading data from socket (stdout or stderr)
    """
    def __init__( self, name, fp, signal, raw_signal, logger = None):
        self.name = name
        self.fp = fp
        flags = fcntl.fcntl(self.fp.fileno(), fcntl.F_GETFL)
        fcntl.fcntl( self.fp.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK )
        self.signal = signal
        self.raw_signal = raw_signal
        self.logger = None
        self.saved = ''
        notifier.socket_add( fp, self._handle_input )
        if logger:
            logger = '%s-%s.log' % ( logger, name )
            try:
                try:
                    os.unlink(logger)
                except (KeyboardInterrupt, SystemExit), e:
                    sys.exit(0)
                except:
                    pass
                self.logger = open(logger, 'w')
                log.info('logging child to "%s"' % logger)
            except IOError:
                log.warning('Error: Cannot open "%s" for logging' % logger)


    def close( self ):
        """
        Close the IO to the child.
        """
        notifier.socket_remove( self.fp )
        self.fp.close()
        if self.logger:
            self.logger.close()
            self.logger = None


    def flush( self ):
        """
        Read as much as possible and close socket later.
        """
        while self._handle_input( self.fp, True ):
            pass

        
    def _handle_input( self, socket, flushing = False ):
        """
        Handle data input from socket.
        """
        try:
            data = self.fp.read(512*1024)
        except IOError, (errno, msg):
            if errno == 11 and not flushing:
                # Resource temporarily unavailable; if we try to read on a
                # non-blocking descriptor we'll get this message.  If we're
                # being called from flush(), it could be because the child
                # process is dead, in which case this errno will occur but
                # we don't want to return True.
                return True
            data = None
        except ValueError:
            # socket already closed
            return False
            
        if not data:
            log.info('No data on %s for pid %s.' % ( self.name, os.getpid()))
            notifier.socket_remove( self.fp )
            self.fp.close()
            if self.logger:
                self.logger.close()
            # FIXME: we know the child is dead.  Rather than wait for it
            # to be reaped we can call __child_died in Process instance.
            return False

        if self.raw_signal.count():
            self.raw_signal.emit(data)

        if self.signal.count() == 0:
            # Nothing connected to the per-line handler, so need to parse.
            # (And if we're dealing with binary data, it may be important
            # _not to parse.)
            return True

        data  = data.replace('\r', '\n')
        lines = data.split('\n')

        # Only one partial line?
        if len(lines) == 1:
            self.saved += data
            return True

        # Combine saved data and first line, send to app
        if self.logger:
            self.logger.write( self.saved + lines[ 0 ] + '\n' )
        self.signal.emit( self.saved + lines[ 0 ] )
        self.saved = ''

        # There's one or more lines + possibly a partial line
        if lines[ -1 ] != '':
            # The last line is partial, save it for the next time
            self.saved = lines[ -1 ]

            # Send all lines except the last partial line to the app
            for line in lines[ 1 : -1 ]:
                if not line:
                    continue
                if self.logger:
                    self.logger.write( line + '\n' )
                self.signal.emit( line )
        else:
            # Send all lines to the app
            for line in lines[ 1 : ]:
                if not line:
                    continue
                if self.logger:
                    self.logger.write( line + '\n' )
                self.signal.emit( line )
        return True


class Watcher(object):
    def __init__( self ):
        log.info('new process watcher instance')
        self.__processes = {}
        self.__timer = None
        self.status = 'running'

    def append( self, proc, cb ):
        self.__processes[ proc ] = cb
        if not self.__timer:
            log.info('start process watching')
            # FIXME: there is no reason to reap 20 times a second.
            self.__timer = notifier.timer_add(50, self.check)


    def check( self ):
        remove_proc = []

        # check all processes
        for p in self.__processes:

            if isinstance( p.child, popen2.Popen3 ):
                pid = p.child.pid
            else:
                pid = p.pid

            try:
                pid, status = os.waitpid( pid, os.WNOHANG )
            except OSError, (errno, err):
                # FIXME: sometimes we get ECHILD ("No child processes", errno=10) even
                # though we definitely have children.  This happens with pppd, for
                # example.  Need to figure out why.
                remove_proc.append( (p, pid, -1) )
                continue

            if not pid:
                continue

            log.info('Dead child: %s (%s)' % ( pid, status ))
            if status == -1:
                log.error('error retrieving process information from %d' % p)
            elif os.WIFEXITED( status ) or os.WIFSIGNALED( status ) or \
                     os.WCOREDUMP( status ):
                remove_proc.append( (p, pid, status) )

        # remove dead processes
        for p, pid, status in remove_proc:
            if p in self.__processes:
                # call stopped callback
                callback = self.__processes[p]
                # Delete the callback from the processes list before calling
                # it, since it's possible the callback could call append
                # again.
                del self.__processes[p]
                callback(pid, status)

        # check if this function needs to be called again
        if not self.__processes:
            # no process left, remove timer
            self.__timer = None
            log.info('stop process watching')
            return False

        # return True to be called again
        return True


    def stop_all( self ):
        if self.status != 'running':
            return
        # stop all childs without waiting
        for p in self.__processes.keys():
            p.stop()
        self.status = 'stopping'


    def kill_all( self ):
        # prevent recursion
        if not self.status in ('running', 'stopping'):
            return
        # make sure every child is stopped
        self.stop_all()
        self.status = 'stopped'


# global watcher instance
proclist = Watcher()



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# stdin.py - Read keys from stdin (not complete lines)
# -----------------------------------------------------------------------------
# $Id: stdin.py 3069 2008-02-13 18:43:13Z dmeyer $
#
# -----------------------------------------------------------------------------
# kaa.input - Kaa input subsystem
# Copyright (C) 2005, 2006 Dirk Meyer, Jason Tackaberry
#
# First Edition: Jason Tackaberry <tack@urandom.ca>
# Maintainer:    Jason Tackaberry <tack@urandom.ca>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'signal' ]

# python imports
import sys
import tty
import termios
import os
import atexit

# kaa imports
import kaa

_tc_orig_settings = None
_getch_enabled = False

_keycode_names = {
    "\x1b\x5b\x41": "up",
    "\x1b\x5b\x42": "down",
    "\x1b\x5b\x43": "right",
    "\x1b\x5b\x44": "left",

    "\x1b\x4f\x50": "F1",
    "\x1b\x4f\x51": "F2",
    "\x1b\x4f\x52": "F3",
    "\x1b\x4f\x53": "F4",
    "\x1b\x5b\x31\x35\x7e": "F5",
    "\x1b\x5b\x31\x37\x7e": "F6",
    "\x1b\x5b\x31\x38\x7e": "F7",
    "\x1b\x5b\x31\x39\x7e": "F8",
    "\x1b\x5b\x32\x30\x7e": "F9",
    "\x1b\x5b\x32\x31\x7e": "F10",
    "\x1b\x5b\x32\x33\x7e": "F11",
    "\x1b\x5b\x32\x34\x7e": "F12",

    "\x1b\x5b\x32\x7e": "ins",
    "\x1b\x5b\x33\x7e": "del",
    "\x1b\x4f\x46": "end",
    "\x1b\x4f\x48": "home",
    "\x1b\x1b": "esc",
    "\x09": "tab",
    "\x0a": "enter",
    "\x20": "space",
    "\x7f": "backspace"
}

def getch():
    global _getch_enabled

    if not _getch_enabled:
        getch_enable()
        _getch_enabled = True

    buf = sys.stdin.read(1)
    while buf in ("\x1b", "\x1b\x4f", "\x1b\x5b") or \
          buf[:3] in ("\x1b\x5b\x31", "\x1b\x5b\x32", "\x1b\x5b\x33"):
        buf += sys.stdin.read(1)
        if buf[-1] == "\x7e":
            break

    #print "KEYCODE:"
    #for c in buf:
    #    print "  " +  hex(ord(c))
    code = buf
    #buf = ""
    if code in _keycode_names:
        return _keycode_names[code]
    elif len(code) == 1:
        return code
    else:
        return "??"


def getch_enable():
    global _tc_orig_settings
    _tc_orig_settings = termios.tcgetattr(sys.stdin.fileno())
    tc = termios.tcgetattr(sys.stdin.fileno())
    tc[3] = tc[3] & ~(termios.ICANON | termios.ECHO)
    tc[6][termios.VMIN] = 1
    tc[6][termios.VTIME] = 0
    termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, tc)
    atexit.register(getch_disable)


def getch_disable():
    global _tc_orig_settings
    if _tc_orig_settings != None:
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, _tc_orig_settings)
    os.system("stty echo")


def _handle_stdin_keypress():
    ch = getch()
    kaa.signals["stdin_key_press_event"].emit(ch)
    return True


_dispatcher = kaa.IOMonitor(_handle_stdin_keypress)

def _keypress_signal_changed(s, flag):
    if flag == kaa.Signal.SIGNAL_CONNECTED and s.count() == 1:
        getch_enable()
        _dispatcher.register(sys.stdin)
    elif flag == kaa.Signal.SIGNAL_DISCONNECTED and s.count() == 0:
        getch_disable()
        _dispatcher.unregister()


# init
signal = kaa.Signal(changed_cb = _keypress_signal_changed)
kaa.signals["stdin_key_press_event"] = signal



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# timer.py - Timer classes for the notifier
# -----------------------------------------------------------------------------
# $Id: timer.py 3170 2008-03-08 19:31:19Z dmeyer $
#
# -----------------------------------------------------------------------------
# kaa.notifier - Mainloop and callbacks
# Copyright (C) 2005-2008 Dirk Meyer, Jason Tackaberry, et al.
#
# First Version: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#                Jason Tackaberry <tack@urandom.ca>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Timer', 'WeakTimer', 'OneShotTimer', 'WeakOneShotTimer',
            'AtTimer', 'OneShotAtTimer', 'POLICY_ONCE', 'POLICY_MANY', 
            'POLICY_RESTART' ]

import logging
import datetime

import nf_wrapper as notifier
from thread import MainThreadCallback, is_mainthread
from kaa.weakref import weakref

POLICY_ONCE = 'once'
POLICY_MANY = 'many'
POLICY_RESTART = 'restart'

# get logging object
log = logging.getLogger('notifier')


def timed(interval, timer=None, policy=POLICY_MANY):
    """
    Decorator to call the decorated function in a Timer. When calling the
    function, a timer will be started with the given interval calling that
    function.  The decorated function will be called from the main thread.

    The timer parameter optionally specifies which timer class should be
    used to wrap the function.  kaa.Timer (default) or kaa.WeakTimer will
    repeatedly invoke the decorated function until it returns False.
    kaa.OneShotTimer or kaa.WeakOneShotTimer will invoke the function once,
    delaying it by the specified interval.  (In this case the return value
    of the decorated function is irrelevant.)

    The policy parameter controls how multiple invocations of the decorated
    function should be handled.  By default (POLICY_MANY), each invocation of
    the function will create a new timer, each firing at the specified
    interval.  If policy is POLICY_ONCE, subsequent invocations are ignored
    while the first timer is still active.  If the policy is POLICY_RESTART,
    subsequent invocations will restart the first timer.

    Note that in the case of POLICY_ONCE or POLICY_RESTART, if the timer is
    currently running, any arguments passed to the decorated function on
    subsequent calls will be discarded.
    """
    if not policy in (POLICY_MANY, POLICY_ONCE, POLICY_RESTART):
        raise RunTimeError('Invalid @kaa.timed policy %s' % policy)

    def decorator(func):
        def newfunc(*args, **kwargs):
            if policy == POLICY_MANY:
                # just start the timer
                t = (timer or Timer)(func, *args, **kwargs)
                t.start(interval)
                return True

            # Object to save the timer in; the function itself for non-methods,
            # or the instance object for methods.
            # object to save the timer in
            obj = func
            # name of the attribute in the object
            name = '__kaa_timer_decorator'

            # Try to find out if the function is actually an instance method.
            # The decorator only sees a function object, even for methods, so
            # this kludge compares the code object of newfunc (this wrapper)
            # with the code object of the first argument's attribute of the
            # function's name.  If they're the same, then we must be decorating
            # a method, and we can attach the timer object to the instance
            # instead of the function.
            if args and newfunc.func_code == \
                        getattr(getattr(args[0], func.func_name, None), 'func_code', None):
                obj  = args[0]
                name = '%s__%s' % (name, func.func_name)

            # check current timer
            if getattr(obj, name, None) and getattr(obj, name).active():
                if policy == POLICY_ONCE:
                    # timer already running and not override
                    return False
                # stop old timer
                getattr(obj, name).stop()

            # create new timer, set it to the object and start it
            t = (timer or Timer)(func, *args, **kwargs)
            setattr(obj, name, weakref(t))
            getattr(obj, name).start(interval)
            return True

        newfunc.func_name = func.func_name
        return newfunc

    return decorator



class Timer(notifier.NotifierCallback):

    def __init__(self, callback, *args, **kwargs):
        super(Timer, self).__init__(callback, *args, **kwargs)
        self.restart_when_active = True
        self._interval = None


    def set_resart_when_active(self, restart):
        self.restart_when_active = restart


    def start(self, interval):
        if not is_mainthread():
            return MainThreadCallback(self.start, interval)()

        if self.active():
            if not self.restart_when_active:
                return
            self.unregister()

        self._id = notifier.timer_add(int(interval * 1000), self)
        self._interval = interval


    def stop(self):
        if not is_mainthread():
            return MainThreadCallback(self.stop)()
        self.unregister()


    def unregister(self):
        if self.active():
            notifier.timer_remove(self._id)
            super(Timer, self).unregister()


    def get_interval(self):
        return self._interval


    def __call__(self, *args, **kwargs):
        if not self.active():
            # This happens if previous timer that has been called during the
            # same notifier step has stopped us. The new notifier could
            # should prevent this.
            log.error('calling callback on inactive timer (%s)' % repr(self))
            return False

        return super(Timer, self).__call__(*args, **kwargs)


class OneShotTimer(Timer):
    """
    A Timer that only gets executed once. If the timer is started again
    inside the callback, make sure 'False' is NOT returned or the timer
    will be removed again without being called. To be on tge same side,
    return nothing in such a callback.
    """
    def __call__(self, *args, **kwargs):
        self.unregister()
        super(Timer, self).__call__(*args, **kwargs)
        return False



class WeakTimer(notifier.WeakNotifierCallback, Timer):
    pass

class WeakOneShotTimer(notifier.WeakNotifierCallback, OneShotTimer):
    pass

class OneShotAtTimer(OneShotTimer):
    """
    Timer that will get executed at a time specified with a list
    of hours, minutes and seconds.
    """
    def start(self, hour=range(24), min=range(60), sec=0):
        if not isinstance(hour, (list, tuple)):
            hour = [ hour ]
        if not isinstance(min, (list, tuple)):
            min = [ min ]
        if not isinstance(sec, (list, tuple)):
            sec = [ sec ]

        self._timings = hour, min, sec
        self._last_time = datetime.datetime.now()
        self._schedule_next()


    def _schedule_next(self):
        """
        Internal function to calculate the next callback time and
        schedule it.
        """
        hour, min, sec = self._timings
        now = datetime.datetime.now()
        # Take the later of now or the last scheduled time for purposes of
        # determining the next time.  If we use the current system time
        # instead, we may end up firing a callback twice for a given time,
        # because due to imprecision we may end up here slightly before (a few
        # milliseconds) the scheduled time.
        t = max(self._last_time, now).replace(microsecond = 0)

        next_sec = [ x for x in sec if t.second < x ]
        next_min = [ x for x in min if t.minute < x ]
        next_hour = [ x for x in hour if t.hour < x ]

        if next_sec:
            next = t.replace(second = next_sec[0])
        elif next_min:
            next = t.replace(minute = next_min[0], second = sec[0])
        elif next_hour:
            next = t.replace(hour = next_hour[0], minute = min[0], second = sec[0])
        else:
            tmrw = t + datetime.timedelta(days = 1)
            next = tmrw.replace(hour = hour[0], minute = min[0], second = sec[0])

        delta = next - now
        super(OneShotAtTimer, self).start(delta.seconds + delta.microseconds / 1000000.0)
        self._last_time = next


class AtTimer(OneShotAtTimer):
    """
    Timer that will get executed at a time specified with a list
    of hours, minutes and seconds. The timer will run until the
    callback returns False or 'stop' is called.
    """
    def __call__(self, *args, **kwargs):
        if super(Timer, self).__call__(*args, **kwargs) != False:
            self._schedule_next()



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# xmlconfig.py - xml config to python converter
# -----------------------------------------------------------------------------
# $Id: xmlconfig.py 2719 2007-06-23 20:46:12Z tack $
#
# -----------------------------------------------------------------------------
# Copyright (C) 2006 Dirk Meyer, Jason Tackaberry
#
# First Edition: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'convert' ]

# python imports
import sys
import pprint

# use minidom because it is part of the python distribution.
# pyxml replaces the parser but everything used here should
# work even without it.
from xml.dom import minidom

def get_value(value, type):
    if not value:
        return eval('%s()' % type)
    if type:
        return eval(type)(value)
    if value.lower() == 'true':
        return True
    if value.lower() == 'false':
        return False
    if value.isdigit():
        return int(value)
    return str(value)


def format_content(node):
    s = node.childNodes[0].data.replace('\t', '        ')
    spaces = ''
    while s:
        if s[0] == ' ':
            spaces += ' '
            s = s[1:]
            continue
        if s[0] == '\n':
            spaces = ''
            s = s[1:]
            continue
        break
    return s.replace('\n' + spaces, '\n').strip()


def nodefilter(node, *names):
    return [ n for n in node.childNodes if n.nodeName in names ]


class Parser(object):

    def __init__(self, package):
        self._package = package

    def _get_schema(self, node):
        schema = []
        for child in node.childNodes:
            if hasattr(self, '_parse_%s' % child.nodeName):
                schema.append(child)
        return schema

    
    def parse(self, node, fd, deep=''):
        fd.write('%s(' % node.nodeName.capitalize())
        first = True
        if node.getAttribute('name'):
            fd.write('name=\'%s\'' % node.getAttribute('name'))
            first = False
        for child in nodefilter(node, 'desc'):
            if not first:
                fd.write(', ')
            first = False
            desc = format_content(child).replace('\'', '\\\'')
            if desc.find('\n') > 0:
                desc = deep + desc.replace('\n', '\n' + deep)
                fd.write('desc=\'\'\'\n%s\n%s\'\'\'' % (desc, deep))
            else:
                fd.write('desc=\'%s\'' % desc)
        getattr(self, '_parse_%s' % node.nodeName.lower())(node, fd, deep, first)
    

    def _parse_var(self, node, fd, deep, first):
        default = node.getAttribute('default')
        deftype = node.getAttribute('type')
        if default or deftype:
            if not first:
                fd.write(', ')
            first = False
            default = get_value(default, deftype)
            fd.write('default=%s' % pprint.pformat(default).strip())

        for child in nodefilter(node, 'values'):
            if not first:
                fd.write(', ')
            first = False
            values = []
            for value in nodefilter(child, 'value'):
                content = value.childNodes[0].data
                values.append(get_value(content, value.getAttribute('type')))
            fd.write('type=%s' % pprint.pformat(tuple(values)).strip())
            break
        fd.write(')')

    
    def _parse_config(self, node, fd, deep, first):
        self._parse_group(node, fd, deep, first)
        

    def _parse_group(self, node, fd, deep, first):
        if not first:
            fd.write(', ')
        deep = deep + '  '
        fd.write('schema=[\n\n' + deep)
        for s in self._get_schema(node):
            self.parse(s, fd, deep)
            fd.write(',\n\n' + deep)
        deep = deep[:-2]
        fd.write(']\n' + deep)
        if node.nodeName == 'config':
            fd.write(", module='%s.config'" % self._package)
        fd.write(')')
    
    
    def _parse_list(self, node, fd, deep, first):
        if not first:
            fd.write(', ')
        schema = self._get_schema(node)
        fd.write('schema=')
        if len(schema) > 1:
            deep = deep + '  '
            fd.write('[\n\n' + deep)
        for s in schema:
            self.parse(s, fd, deep)
            if len(schema) > 1:
                fd.write(',\n\n' + deep)
        if len(schema) > 1:
            deep = deep[:-2]
            fd.write(']\n' + deep)
        type = node.getAttribute('type')
        if type and node.nodeName == 'dict':
            fd.write(', type=%s' % type)
        defaults = {}
        for var in nodefilter(node, 'set'):
            value = get_value(var.getAttribute('value'), None)
            defaults[var.getAttribute('key')] = value
        if defaults:
            fd.write(', defaults=%s' % pprint.pformat(defaults).strip())
            
        fd.write(')')


    def _parse_dict(self, node, fd, deep, first):
        self._parse_list(node, fd, deep, first)
        


def convert(xml, python, package):
    tree = minidom.parse(xml).firstChild
    if tree.nodeName != 'config':
        raise RuntimeError('%s is no valid cxml file' % xml)
    
    # out = sys.__stdout__
    out = open(python, 'w')
    
    out.write('# auto generated file\n\n')
    out.write('from kaa.config import Var, Group, Dict, List, Config\n\n')
    out.write('config = ')

    Parser(package).parse(tree, out)
    out.write('\n\n')
    for child in tree.childNodes:
        if child.nodeName != 'code':
            continue
        out.write(format_content(child) + '\n\n')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Andreas Büsching <crunchy@bitkipper.net>
#
# a generic dispatcher implementation
#
# $Id: dispatch.py 2673 2007-05-08 22:38:50Z tack $
#
# Copyright (C) 2006
#	Andreas Büsching <crunchy@bitkipper.net>
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""generic implementation of external dispatchers, integratable into
several notifiers."""

from copy import copy

# required for dispatcher use
MIN_TIMER = 100

__dispatchers = []

def dispatcher_add( method ):
    """The notifier supports external dispatcher functions that will be called
    within each scheduler step. This functionality may be usful for
    applications having an own event mechanism that needs to be triggered as
    often as possible. This method registers a new dispatcher function. To
    ensure that the notifier loop does not suspend to long in the sleep state
    during the select a minimal timer MIN_TIMER is set to guarantee that the
    dispatcher functions are called at least every MIN_TIMER milliseconds."""
    global __dispatchers
    __dispatchers.append( method )

def dispatcher_remove( method ):
    """Removes an external dispatcher function from the list"""
    global __dispatchers
    if method in __dispatchers:
        __dispatchers.remove( method )

def dispatcher_run():
    global __dispatchers
    for disp in copy( __dispatchers ):
        if not disp():
            dispatcher_remove( disp )

def dispatcher_count():
    return len(__dispatchers)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# svn2log.py - create ChangeLog file based on svn log
# -----------------------------------------------------------------------------
# $Id: svn2log.py 2597 2007-03-23 18:48:05Z dmeyer $
#
# -----------------------------------------------------------------------------
# Copyright (C) 2007 Dirk Meyer, Jason Tackaberry
#
# First Edition: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

# python imports
import sys
import os
import textwrap
import popen2
import re
import xml.sax

try:
    from xmlutils import SaxTreeHandler
except ImportError:
    from kaa.xmlutils import SaxTreeHandler
    
class Entry(object):
    def __init__(self, author, date):
        self.author = author
        self.date = date
        self.changes = []

    def write(self, writer):
        writer.write('%s  %s\n' % (self.date, self.author))
        for revision, msg, files, changed_listing in self.changes:
            if not changed_listing:
                files = textwrap.wrap(', '.join(files), width=70)
            if len(files) == 1 and len(files[0]) + len(msg) < 70:
                writer.write('\n\t* %s: %s\n' % (files[0], msg))
                continue
            for line in files:
                writer.write('\n\t* %s' % line)
            writer.write(':\n')
            for delimiter in ('o ', '- ', '-'):
                found = 0
                for l in msg.split('\n'):
                    if l.startswith(delimiter):
                        found += 1
                if found > 1:
                    p = ''
                    for l in msg.split('\n'):
                        if not l.startswith(delimiter) or not p:
                            p += l
                            continue
                        if p.startswith(delimiter):
                            p = p[len(delimiter):].lstrip()
                        writer.write('\t' + '\n\t'.join(textwrap.wrap(p)) + '\n\n')
                        p = l
                    p = p[len(delimiter):].lstrip()
                    writer.write('\t' + '\n\t'.join(textwrap.wrap(p)))
                    break
            else:
                writer.write('\t' + '\n\t'.join(textwrap.wrap(msg)))
            writer.write('\n')
        writer.write('\n')
        
class LogParser(SaxTreeHandler):

    def __init__(self, writer, prefix, user):
        SaxTreeHandler.__init__(self, 'logentry')
        prefix = '/(trunk|branches/[^/]+)(/WIP)?/(%s)' % '|'.join(prefix)
        self._prefix = re.compile(prefix)
        self._entry = None
        self._user = user
        self._writer = writer
        
    def handle(self, node):
        revision = node.getattr('revision')
        date = msg = author = ''
        files = []
        changed_listing = False
        for c in node.children:
            c.content = c.content.encode('latin-1', 'ignore')
            if c.name == 'msg':
                msg = re.subn('  ', ' ', c.content)[0]
            if c.name == 'date':
                date = c.content[:c.content.find('T')]
            if c.name == 'author':
                author = c.content
                if author in self._user:
                    author = self._user[author]
                else:
                    print 'unknown author', author
            if c.name == 'paths':
                for path in c.children:
                    if self._prefix.search(path.content):
                        f = self._prefix.sub('', path.content)
                        if not len(f):
                            f = path.content
                        if path.getattr('action') == "D":
                            files.append(f[1:] + " (removed)")
                            changed_listing = True
                        elif path.getattr('action') == "A":
                            files.append(f[1:] + " (added)")
                            changed_listing = True
                        else:
                            files.append(f[1:])
                if not len(files):
                    print 'error detecting files'
                    for path in c.children:
                        print path.content
                    print
                    
        if self._entry and (self._entry.author != author or self._entry.date != date):
            self._entry.write(self._writer)
            self._entry = None
        if not self._entry:
            self._entry = Entry(author, date)
        self._entry.changes.append((revision, msg, files, changed_listing))

    def finalize(self):
        self._entry.write(self._writer)


def svn2log(module):
    if not os.path.isfile('ChangeLog.in') or not os.path.isdir('.svn'):
        return
    prefix = [ module ]
    users = {}
    for line in open('ChangeLog.in'):
        if line.startswith('user'):
            login, name = line[4:].strip().split(' ', 1)
            users[login] = name
        if line.startswith('aka'):
            prefix.append(line[4:].strip())
    # make sure that module2 is higher in the last than module itself
    prefix.sort()
    prefix.reverse()
    
    # Create a parser
    parser = xml.sax.make_parser()

    reader = popen2.popen2('svn log -v --xml')[0]
    writer = open('ChangeLog', 'w')

    dh = LogParser(writer, prefix, users)

    # Tell the parser to use our handler
    parser.setContentHandler(dh)

    # parse the input, add file:// so the parser will find
    # the dtd for program files.
    parser.parse(reader)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# __init__.py - Interface to kaa.notifier
# -----------------------------------------------------------------------------
# $Id: __init__.py 3208 2008-03-20 15:46:27Z dmeyer $
#
# -----------------------------------------------------------------------------
# kaa.notifier - Mainloop and callbacks
# Copyright (C) 2005-2008 Dirk Meyer, Jason Tackaberry, et al.
#
# First Version: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#                Jason Tackaberry <tack@urandom.ca>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

# Import all classes, functions and decorators that are part of the API

# Callback classes
from callback import Callback, WeakCallback

# Signal and dict of Signals
from signals import Signal, Signals

# InProgress class
from async import TimeoutException, InProgress, InProgressCallback, InProgressSignals

# Thread callbacks, helper functions and decorators
from thread import MainThreadCallback, NamedThreadCallback, ThreadCallback, \
     is_mainthread, threaded, MAINTHREAD, synchronized

# Timer classes and decorators
from timer import Timer, WeakTimer, OneShotTimer, WeakOneShotTimer, AtTimer, \
     OneShotAtTimer, timed, POLICY_ONCE, POLICY_MANY, POLICY_RESTART

# IO/Socket handling
from sockets import IOMonitor, WeakIOMonitor, Socket, IO_READ, IO_WRITE

# Event and event handler classes
from event import Event, EventHandler, WeakEventHandler

# coroutine decorator and helper classes
from coroutine import NotFinished, coroutine

# process management
from popen import Process

# special gobject thread support
from gobject import GOBJECT, gobject_set_threaded



from kaa import timed, Timer, OneShotTimer, main, \
     is_mainthread, threaded, MAINTHREAD, POLICY_ONCE, POLICY_RESTART


class Foo(object):

    @timed(0.1, policy=POLICY_ONCE)
    def poll(self, x):
        if not x:
            return False
        y = x.pop(0)
        print '1', y
        return True

    @timed(0.1, policy=POLICY_ONCE)
    def poll2(self, x):
        if not x:
            return False
        y = x.pop(0)
        print '2', y
        return True

    @threaded('name')
    def foo(self):
        print 'foo is mainthread:', is_mainthread()
        self.bar()
        
    @threaded(MAINTHREAD)
    def bar(self):
        print 'bar is mainthread:', is_mainthread()
        

@timed(0.1, policy=POLICY_RESTART)
def poll(x):
    if not x:
        return False
    y = x.pop(0)
    print y
    return True


@timed(0.1, OneShotTimer, policy=POLICY_RESTART)
def bla(f, msg):
    print msg
    f.foo()
    f.bar()
    
f = Foo()
f.poll([0,1,2,3,4,5])
f.poll2(['a','b','c','d','e','f'])

poll([10,11,12,13,14,15])
bla(f, 'test')
bla(f, 'test2')
OneShotTimer(poll, [20,21,22,23,24]).start(0.3)
OneShotTimer(f.poll, [30,31,32,33,34]).start(0.3)
main.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Andreas Büsching <crunchy@bitkipper.net>
#
# log - a logging facility for the generic notifier module
#
# Copyright (C) 2005, 2006
#		Andreas Büsching <crunchy@bitkipper.net>
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

import logging
import sys

instance = logging.getLogger( 'notifier' )
instance.addHandler( logging.StreamHandler( sys.stderr ) )

debug = instance.debug
info = instance.info
warn = instance.warn
error = instance.error
critical = instance.critical
exception = instance.exception



import sys
import kaa

# method can be either
# 0: use thread based notifier
# 1: use twisted thread based notifier
# 2: use twisted experimental pynotifier
#    Known Bug: Twisted does not stop the reactor on SystemExit. If a
#    callback calls sys.exit() to shutdown the program, this won't work.
#    This has to be fixed in pynotifier
method = 2

# test auto-stop or stop with C-c
stop_on_exit = True

# get reactor
from twisted.internet import reactor

def twisted_callback1():
    print "twisted", kaa.is_mainthread()
    
def twisted_callback2():
    print "twisted (shutdown)", kaa.is_mainthread()
    if not stop_on_exit:
        return
    if method == 2:
        kaa.main.stop()
    else:
        reactor.stop()
    
def kaa_callback():
    print 'kaa', kaa.is_mainthread()

def shutdown_callback():
    print 'shutdown signal'

if method == 0:
    # select the thread loop and provide callbacks for a hander and how to
    # stop the real mainloop
    loop = kaa.main.select_notifier('thread', handler = reactor.callFromThread, \
                                    shutdown = reactor.stop)
    # stop the thread when twisted is done. Without this code the app will wait
    # until the nf thread will be done which will never happen.
    reactor.addSystemEventTrigger('after', 'shutdown', loop.stop)

if method == 1:
    # there is special code in kaa that does the same as method 0
    kaa.main.select_notifier('twisted')
if method == 2:
    # or use the real twisted notifier
    kaa.main.select_notifier('twisted_experimental')

reactor.callLater(0.5, twisted_callback1)
reactor.callLater(1.5, twisted_callback2)
kaa.Timer(kaa_callback).start(1)
kaa.main.signals['shutdown'].connect(shutdown_callback)

if method == 2:
    # the twisted_experimental needs kaa.main.run()
    kaa.main.run()
else:
    reactor.run()

print 'done'



#!/usr/bin/env python
#
# kaaconf-editor - GTK-based kaa.config file editor
# 
# 
#  Still TODO:
#    - Support List and Dict types
#    - Implement filter
#    - Allow specifying module to import when config file doesn't contain it
#      (also needed for creating new config files)
#    - Miscellaneous polish
#    - Code cleanup and commenting

import os
import sys

import pygtk
pygtk.require('2.0')
import gtk
import gtk.glade
import gobject

import kaa, kaa.config

# use gtk main loop
kaa.main.select_notifier('gtk')

class ProxyCellRenderer(gtk.CellRenderer):

    __gproperties__ = {
        'value' : (gobject.TYPE_PYOBJECT, None, None, gobject.PARAM_READWRITE)
        }

    __gsignals__ = {
        'changed': (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, (gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)) 
    }

    def __init__(self, widget = None):
        self.__gobject_init__()
        self._value = None
        self._text_renderer = gtk.CellRendererText()
        self._text_renderer.set_property('editable', True)
        self._text_renderer.connect('edited', self.edited_handler)

        self._toggle_renderer = gtk.CellRendererToggle()
        self._toggle_renderer.set_property('xalign', 0.0)
        self._toggle_renderer.set_property('activatable', True)

        self._combo_renderer = gtk.CellRendererCombo()
        self._combo_renderer.set_property('editable', True)
        self._combo_renderer.set_property('has-entry', False)
        self._combo_renderer.set_property('text-column', 0)
        self._combo_renderer.connect('edited', self.edited_handler)

        self._cur_renderer = None

    def do_get_property(self, property):
        if property.name == 'value':
            return self._value
        else:
            raise AttributeError, 'unknown property %s' % property.name


    def do_set_property(self, property, value):
        if property.name != 'value':
            raise AttributeError, 'unknown property %s' % property.name

        self._value = value
        if self._value[0] == 'text':
            self._cur_renderer, mode = self._text_renderer, gtk.CELL_RENDERER_MODE_EDITABLE
            self._text_renderer.set_property('text', self._value[1])
        elif self._value[0] == 'checkbox':
            self._cur_renderer, mode = self._toggle_renderer, gtk.CELL_RENDERER_MODE_ACTIVATABLE
            self._toggle_renderer.set_property('active', self._value[1])
        elif self._value[0] == 'combo':
            self._cur_renderer, mode = self._combo_renderer, gtk.CELL_RENDERER_MODE_EDITABLE
            self._combo_renderer.set_property('text', self._value[1])
            self._combo_renderer.set_property('model', self._value[2])

        self.set_property('mode', mode)


    def do_render(self, window, widget, background_area, cell_area, expose_area, flags):
        if self._value == None:
            self._text_renderer.set_property('text', '<no value>')

        return self._cur_renderer.render(window, widget, background_area, cell_area, expose_area, flags)

    def do_get_size(self, widget, cell_area):
        return self._cur_renderer.get_size(widget, cell_area)

    def do_activate(self, event, widget, path, background_area, cell_area, flags):
        if self._value[0] == 'checkbox':
            self._value[1] = not self._value[1]
            if not event:
                # When spacebar is used to toggle, event is None.  I have to
                # explicitly queue a redraw here or else it doesn't reflect the
                # toggle.  Why?
                widget.queue_draw()
            self.emit('changed', path, self._value[1])
        return True


    def do_start_editing(self, event, widget, path, background_area, cell_area, flags):
        if not event:
            # Complains when no event, so create dummy
            event = gtk.gdk.Event(gtk.gdk.NOTHING)
        self._edit_data = widget, path, self._value
        return self._cur_renderer.start_editing(event, widget, path, background_area, cell_area, flags)


    def edited_handler(self, entry, cell, text):
        tree, path, value = self._edit_data
        value[1] = text
        self.set_property('value', value)
        # Combo widget loses focus, refocus tree after edit.
        # FIXME: doesn't work when edit is cancelled
        tree.grab_focus()
        self._edit_data = None
        self.emit('changed', path, text)

gobject.type_register(ProxyCellRenderer)



class Editor(object):
    def __init__(self):
        self._current_group = None
        self._current_cfg_path = [], []
        self._filter_changed_timer = kaa.OneShotTimer(self.filter_changed_timeout_handler)

        gladefile = os.path.dirname(os.path.abspath(__file__)) + '/kaaconf-editor.glade'
        self.xml = gtk.glade.XML(gladefile, 'main')
        handlers = {
            'on_toolbar_open_clicked': self.open_handler,
            'on_toolbar_save_clicked': self.save_handler,
            'on_toolbar_quit_clicked': self.exit_handler,
            'on_main_delete_event': self.exit_handler,
            'on_entry_filter_changed': self.filter_changed_handler,
            'on_button_clear_clicked': self.button_clear_handler,
        }
        self.xml.signal_autoconnect(handlers)


        grouptree = self.xml.get_widget('tree-groups')
        grouptree.connect('row-activated', self.tree_row_activated_handler)
        grouptree.connect('cursor-changed', self.tree_cursor_changed_handler)
        treestore = gtk.TreeStore(gtk.gdk.Pixbuf, str)
        grouptree.set_model(treestore)

        col = gtk.TreeViewColumn()
        cr = gtk.CellRendererPixbuf()
        col.pack_start(cr, expand = False)
        col.add_attribute(cr, 'pixbuf', 0)

        cr = gtk.CellRendererText()
        col.pack_start(cr)
        col.add_attribute(cr, 'text', 1)
        grouptree.append_column(col)

        grouptree.set_search_column(1)

        vartree = self.xml.get_widget('tree-variables')
        vartree.connect('cursor-changed', self.list_cursor_changed_handler)
        liststore = gtk.ListStore(str, gobject.TYPE_PYOBJECT)
        vartree.set_model(liststore)
        cr = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Name', cr, text = 0)
        col.set_reorderable(True)
        col.set_sort_column_id(0)
        col.set_min_width(150)
        col.set_resizable(True)
        vartree.append_column(col)

        cr = ProxyCellRenderer()
        cr.connect('changed', self.variable_changed)
        col = gtk.TreeViewColumn('Value', cr, value = 1)
        vartree.append_column(col)

        self._group_tree = grouptree
        self._var_tree = vartree


    def open(self, filename):
        try:
            config = kaa.config.get_config(filename)
        except Exception, msg:
            alert = gtk.MessageDialog(buttons = gtk.BUTTONS_CLOSE)
            alert.set_markup('<span weight="bold" size="larger">Open failed</span>\n\n%s' % msg)
            alert.run()
            alert.destroy()
            return
        
        self._config = config
        self.populate_groups()


    def populate_groups(self):
        store = self._group_tree.get_model()
        store.clear()
        pb = self.xml.get_widget('main').render_icon(gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU)

        def add_group(parent, group):
            for var in group.get_variables():
                if isinstance(getattr(group, var), kaa.config.Group):
                    node = store.append(parent, [pb, var])
                    add_group(node, getattr(group, var))

        root = store.append(None, [pb, '/'])
        add_group(root, self._config)

        self._group_tree.expand_all()
        self._group_tree.set_cursor((0,))


    def populate_variables_from_cfgpath(self):
        path = list(self._current_cfg_path[0])
        group = self._config
        while path:
            group = getattr(group, path.pop(0))

        self._current_group = group
        self._current_cfg_path = self._current_cfg_path[0], []
        store = self._var_tree.get_model()
        store.clear()
        for var in group.get_variables():
            proxy = getattr(group, var)
            if isinstance(proxy, kaa.config.Group):
                continue

            vartype = kaa.config.get_type(proxy)
            if isinstance(vartype, (tuple, list)):
                model = gtk.ListStore(str)
                for option in vartype:
                    model.append([option])
                store.append([var, ['combo', str(proxy), model]])
            elif vartype == bool:
                store.append([var, ['checkbox', bool(proxy)]])
            elif vartype in (kaa.config.List, kaa.config.Dict):
                store.append([var, ['text', '[Dict type not yet implemented]']])
                
            else:
                store.append([var, ['text', str(proxy)]])

        self._var_tree.columns_autosize()

    def variable_changed(self, renderer, path, value):
        model = self._var_tree.get_model()
        iter = model.get_iter(path)
        current_var_name = model.get_value(iter, 0)
        setattr(self._current_group, current_var_name, value)


    def tree_row_activated_handler(self, tree, path, column):
        if tree.row_expanded(path):
            tree.collapse_row(path)
        else:
            tree.expand_row(path, False)

    def tree_cursor_changed_handler(self, tree):
        self._current_cfg_path = self._get_current_cfg_path()
        self.populate_variables_from_cfgpath()
        self.set_info()
        self.update_status()


    def list_cursor_changed_handler(self, tree):
        self._current_cfg_path = self._get_current_cfg_path()
        self.set_info()
        self.update_status()


    def _get_current_cfg_path(self):
        path, col = self._group_tree.get_cursor()
        model = self._group_tree.get_model()
        iter = model.get_iter(path)
        parts = []
        while iter:
            next_iter = model.iter_parent(iter)
            if not next_iter:
                break

            part = model.get_value(iter, 1)
            parts.insert(0, part)
            iter = next_iter

        model = self._var_tree.get_model()
        path, col = self._var_tree.get_cursor()
        if path:
            iter = model.get_iter(path)
            return parts, [model.get_value(iter, 0)]
        return parts, []


    def set_info(self):
        if not self._current_cfg_path[1]:
            var = self._current_group
        else:
            var = getattr(self._current_group, self._current_cfg_path[1][0])

        path = self._current_cfg_path[0] + self._current_cfg_path[1]
        vartype = kaa.config.get_type(var)
        if isinstance(vartype, (tuple, list)):
            # XXX: do we want to list them?
            type_name = 'Enumerated Type'
        else:
            typemap = {
                kaa.config.Group: 'Group',
                kaa.config.Config: 'Group',
                kaa.config.List: 'List',
                kaa.config.Dict: 'Dict',
                bool: 'Boolean',
                int: 'Integer',
                str: 'String',
                unicode: 'Unicode String'
            }
            type_name = typemap.get(vartype, 'Unknown')

        default = kaa.config.get_default(var)
        if not default:
            default = '<None>'

        self.xml.get_widget('label-name').set_label('.'.join(path))
        self.xml.get_widget('label-description').set_label(kaa.config.get_description(var))
        self.xml.get_widget('label-type').set_label(type_name)
        self.xml.get_widget('label-default').set_label(str(default))
        

    def update_status(self):
        path = self._current_cfg_path[0] + self._current_cfg_path[1]
        status = self.xml.get_widget('statusbar')
        status.push(0, '.'.join(path))


    def open_handler(self, *args):
        chooser = gtk.FileChooserDialog(title = 'Select Config File', action = gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                                   gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        chooser.run()
        filename = chooser.get_filename()
        chooser.destroy()

        self.open(filename)


    def exit_handler(self, *args):
        sys.exit(0)


    def save_handler(self, *args):
        self._config.save(force = True)


    def filter_changed_handler(self, text):
        self._filter_changed_timer.start(0.5)


    def filter_changed_timeout_handler(self):
        print "Filter timeout (NYI)"
 

    def button_clear_handler(self, *args):
        self.xml.get_widget('entry-filter').set_text('')
        self._filter_changed_timer.stop()
        self.filter_changed_timeout_handler()


editor = Editor()
if len(sys.argv) > 1:
    editor.open(sys.argv[1])

kaa.main.run()




# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# core.py - distribution core functions for kaa packages
# -----------------------------------------------------------------------------
# $Id: core.py 3206 2008-03-20 14:16:48Z dmeyer $
#
# -----------------------------------------------------------------------------
# Copyright (C) 2006 Dirk Meyer, Jason Tackaberry
#
# First Edition: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

# python imports
import os
import sys
import math
import stat
import re
import tempfile
import time
import distutils.core
import distutils.sysconfig

# internal imports
from version import Version
try:
    from build_py import build_py
    from svn2log import svn2log
except ImportError, e:
    # Small hack before the next release. FIXME: this should be
    # removed for future versions again.
    print 'Error: detected files from previous kaa.base version!'
    print 'Please reinstall kaa by deleting all \'build\' subdirs'
    print 'for all kaa source modules and delete the kaa directory'
    print 'in the site-package directory'
    raise e

__all__ = ['compile', 'check_library', 'get_library', 'setup', 'ConfigFile',
           'Extension', 'Library']

_libraries = {}

def compile(includes, code='', args=''):
    fd, outfile = tempfile.mkstemp()
    os.close(fd)
    args += ' %s %s' % (os.getenv('CFLAGS', ''), os.getenv('LDFLAGS', ''))
    cc = os.getenv('CC', 'cc')
    f = os.popen("%s -x c - -o %s %s 2>/dev/null >/dev/null" % (cc, outfile, args), "w")
    if not f:
        return False

    for i in includes:
        f.write('#include %s\n' % i)
    f.write('int main() { ' + code + '\nreturn 0;\n};\n')
    result = f.close()

    if os.path.exists(outfile):
        os.unlink(outfile)

    if result == None:
        return True
    return False


def check_library(name, *args, **kwargs):
    lib = Library(name)
    if len(args) < 2 and not kwargs:
        minver = args[0]
        print 'checking for', name, '>=', minver, '...',
        sys.__stdout__.flush()
        found, version = lib.check(minver)
        if found:
            print version
        elif version:
            print 'no (%s)' % version
            return
        else:
            print 'no'
            return
    else:
        for var in ('include_dirs', 'library_dirs', 'libraries'):
            if var in kwargs:
                setattr(lib, var, kwargs.pop(var))
        print 'checking for', name, '...',
        sys.__stdout__.flush()
        if lib.compile(*args, **kwargs):
            print 'ok'
        else:
            print 'no'
            return
    _libraries[name] = lib
    return lib


def get_library(name):
    lib = _libraries.get(name)
    if lib and lib.valid:
        return lib
    return None


class Library(object):
    def __init__(self, name):
        self.name = name
        self.include_dirs = []
        self.library_dirs = []
        self.libraries = []
        self.version = None
        self.valid = False


    def compare_versions(self, a, b):
        # Get maximum component length in both version.
        maxlen =  max([ len(x) for x in (a + '.' + b).split('.') ])
        # Pad each component of A and B to maxlen
        a = [ x.zfill(maxlen) for x in a.split('.') ]
        b = [ x.zfill(maxlen) for x in b.split('.') ]
        # TODO: special handling of rc and beta (others?) suffixes, so
        # that 1.0.0 > 1.0.0rc1 
        return cmp(a, b)


    def get_numeric_version(self, version = None):
        """
        Returns an integer version suitable for numeric comparison.  This
        returns the version of the library gotten after calling check().  If
        check() was not called this method will raise an exception.

        This algorithm is pretty naive and won't work at all if the version
        contains any non-numeric characters.
        """
        if version is None:
            version = self.version

        if version is None:
            raise ValueError, "Version is not known; call check() first"

        if not version.replace(".", "").isdigit():
            raise ValueError, "Version cannot have non-numeric characters."

        # Performs: 1.2.3.4 => 4<<0 + 3<<9 + 2<<18 + 1<<27
        return sum([ int(v) << 9*s for s,v in enumerate(reversed(version.split('.'))) ])


    def check(self, minver):
        """
        Check dependencies add add the flags to include_dirs, library_dirs and
        libraries. The basic logic is taken from pygame.
        """
        if os.system("%s-config --version >/dev/null 2>/dev/null" % self.name) == 0:
            # Use foo-config if it exists.
            command = "%s-config %%s 2>/dev/null" % self.name
            version_arg = "--version"
        elif os.system("pkg-config %s --exists >/dev/null 2>/dev/null" % self.name) == 0:
            # Otherwise try pkg-config foo.
            command = "pkg-config %s %%s 2>/dev/null" % self.name
            version_arg = "--modversion"
        else:
            return False, None

        version = os.popen(command % version_arg).read().strip()
        if len(version) == 0:
            return False, None
        if minver and self.compare_versions(minver, version) > 0:
            return False, version

        for inc in os.popen(command % "--cflags").read().strip().split(' '):
            if inc[2:] and not inc[2:] in self.include_dirs:
                self.include_dirs.append(inc[2:])

        for flag in os.popen(command % "--libs").read().strip().split(' '):
            if flag[:2] == '-L' and not flag[2:] in self.library_dirs:
                self.library_dirs.append(flag[2:])
            if flag[:2] == '-l' and not flag[2:] in self.libraries:
                self.libraries.append(flag[2:])

        self.version = version
        self.valid = True
        return True, version


    def compile(self, includes, code='', args='', extra_libraries = []):
        ext_args = [ "-L%s" % x for x in self.library_dirs ]
        ext_args += [ "-I%s" % x for x in self.include_dirs ]

        for lib in extra_libraries:
            if isinstance(lib, basestring):
                lib = get_library(lib)
            for dir in lib.library_dirs:
                ext_args.append("-L%s" % dir)
            for dir in lib.include_dirs:
                ext_args.append("-I%s" % dir)

        # FIXME: we do not check if the lib is installed, we only
        # check the header files.
        if compile(includes, code, args + ' %s' % ' '.join(ext_args)):
            self.valid = True
            return True

        return False



class ConfigFile(object):
    """
    Config file for the build process.
    """
    def __init__(self, filename):
        self.file = os.path.abspath(filename)
        # create config file
        open(self.file, 'w').close()


    def append(self, line):
        """
        Append something to the config file.
        """
        f = open(self.file, 'a')
        f.write(line + '\n')
        f.close()


    def define(self, variable, value=None):
        """
        Set a #define.
        """
        if value == None:
            self.append('#define %s' % variable)
        else:
            self.append('#define %s %s' % (variable, value))


    def unlink(self):
        """
        Delete config file.
        """
        os.unlink(self.file)


class Extension(object):
    """
    Extension wrapper with additional functions to find libraries and
    support for config files.
    """
    def __init__(self, output, files, include_dirs=[], library_dirs=[],
                 libraries=[], extra_compile_args = [], config=None):
        """
        Init the Extension object.
        """
        self.output = output
        self.files = files
        self.include_dirs = include_dirs[:]
        self.library_dirs = library_dirs[:]
        self.libraries = libraries[:]
        self.library_objects = []
        self.extra_compile_args = ["-Wall"] + extra_compile_args
        if config:
            self.configfile = ConfigFile(config)
        else:
            self.configfile = None


    def config(self, line):
        """
        Write a line to the config file.
        """
        if not self.configfile:
            raise AttributeError('No config file defined')
        self.configfile.append(line)


    def add_library(self, name):
        """
        """
        lib = get_library(name)
        if lib and lib not in self.library_objects:
            self.library_objects.append(lib)
        return False


    def get_library(self, name):
        for lib in self.library_objects:
            if lib.name == name:
                return lib


    def check_library(self, name, minver):
        """
        Check dependencies add add the flags to include_dirs, library_dirs and
        libraries. The basic logic is taken from pygame.
        """
        try:
            lib = Library(name)
            found, version = lib.check(minver)
            if not found:
                if not version:
                    version = 'none'
                raise ValueError, 'requires %s version %s (none found)' % \
                                  (name, minver, version)
            self.library_objects.append(lib)
            return True
        except Exception, e:
            self.error = e

        return False


    def check_cc(self, includes, code='', args='', extra_libraries = []):
        """
        Check the given code with the linker. The optional parameter args
        can contain additional command line options like -l.
        """
        ext_args = []
        for lib in self.library_objects + extra_libraries:
            for dir in lib.library_dirs:
                ext_args.append("-L%s" % dir)
            for dir in lib.include_dirs:
                ext_args.append("-I%s" % dir)
        args += ' %s' % ' '.join(ext_args)
        return compile(includes, code, args)


    def has_python_h(self):
        """
        Return True if Python.h is found in the system. If not compiling
        the extention will fail in most cases.
        """
        return os.path.exists(os.path.join(distutils.sysconfig.get_python_inc(), 'Python.h'))

        
    def convert(self):
        """
        Convert Extension into a distutils.core.Extension.
        """
        library_dirs = self.library_dirs[:]
        include_dirs = self.include_dirs[:]
        libraries = self.libraries[:]
        for lib in self.library_objects:
            library_dirs.extend(lib.library_dirs)
            include_dirs.extend(lib.include_dirs)
            libraries.extend(lib.libraries)

        return distutils.core.Extension(self.output, self.files,
                                        library_dirs=library_dirs,
                                        include_dirs=include_dirs,
                                        libraries=libraries,
                                        extra_compile_args=self.extra_compile_args)

    def __del__(self):
        """
        Delete the config file.
        """
        if self.configfile:
            self.configfile.unlink()


class EmptyExtensionsList(list):
    """
    A list that is non-zero even when empty.  Used for the ext_modules
    kwarg in setup() for modules with no ext_modules.

    This is a kludge to solve a peculiar problem.  On architectures like
    x86_64, distutils will install "pure" modules (i.e. no C extensions)
    under /usr/lib, and platform-specific modules (with extensions) under
    /usr/lib64.  This is a problem for kaa, because with kaa we have a single
    namespace (kaa/) that have several independent modules: that is, each
    module in Kaa can be installed separately, but they coexist under the
    kaa/ directory hierarchy.

    On x86_64, this results in some kaa modules being installed in /usr/lib
    and others installed in /usr/lib64.  The first problem is that
    kaa/__init__.py is provided by kaa.base, so if kaa.base is installed
    in /usr/lib/ then /usr/lib64/python*/site-packages/kaa/__init__.py does
    not exist and therefore we can't import any modules from that.  We could
    drop a dummy __init__.py there, but then python will have two valid kaa
    modules and only ever see one of them, the other will be ignored.

    So this kludge makes distutils always think the module has extensions,
    so it will install in the platform-specific libdir.  As a result, on
    x86_64, all kaa modules will be installed in /usr/lib64, which is what
    we want.
    """
    def __nonzero__(self):
        return True




def setup(**kwargs):
    """
    A setup script wrapper for kaa modules.
    """
    def _find_packages((kwargs, prefix), dirname, files):
        """
        Helper function to create 'packages' and 'package_dir'.
        """
        for key, value in kwargs.get('plugins', {}).items():
            if dirname.startswith(value):
                python_dirname = key + dirname[len(value):].replace('/', '.')
                break
        else:
            python_dirname = prefix + dirname[3:].replace('/', '.')
            # Anything under module/src/extensions/foo gets translated to 
            # kaa.module.foo.
            python_dirname = python_dirname.replace(".extensions.", ".")
        if '__init__.py' in files or python_dirname.endswith('plugins'):
            kwargs['package_dir'][python_dirname] = dirname
            kwargs['packages'].append(python_dirname)


    if 'module' not in kwargs:
        raise AttributeError('\'module\' not defined')

    # create name
    kwargs['name'] = 'kaa-' + kwargs['module']

    # search for source files and add it package_dir and packages
    kwargs['package_dir'] = {}
    kwargs['packages']    = []
    if kwargs['module'] == 'base':
        os.path.walk('src', _find_packages, (kwargs, 'kaa'))
    else:
        os.path.walk('src', _find_packages, (kwargs, 'kaa.' + kwargs['module']))

    if 'plugins' in kwargs:
        del kwargs['plugins']
        
    # convert Extensions
    if kwargs.get('ext_modules'):
        kaa_ext_modules = kwargs['ext_modules']
        ext_modules = []
        for ext in kaa_ext_modules:
            ext_modules.append(ext.convert())
        kwargs['ext_modules'] = ext_modules
        # We need to compile an extension, so first do a naive check to ensure
        # Python headers are available.
        if not os.path.exists(os.path.join(distutils.sysconfig.get_python_inc(), 'Python.h')):
            print "- Python headers not found; please install python development package."
            sys.exit(1)

    else:
        # No extensions, but trick distutils into thinking we do have, so
        # the module gets installed in the platform-specific libdir.
        kwargs['ext_modules'] = EmptyExtensionsList()

    # check version.py information
    write_version = False
    if 'version' in kwargs:
        write_version = True
        # check if a version.py is there
        if os.path.isfile('src/version.py'):
            # read the file to find the old version information
            f = open('src/version.py')
            for line in f.readlines():
                if line.startswith('VERSION'):
                    if eval(line[10:-1]) == kwargs['version']:
                        write_version = False
                    break
            f.close()
           
    if len(sys.argv) > 1 and sys.argv[1] in ('bdist_rpm', 'sdist') and \
           os.path.isfile('ChangeLog.in'):
        # FIXME: find a better way to detect if we need to create a
        # ChangeLog file or not.
        print 'generate ChangeLog'
        svn2log(kwargs['module'])
        
    # delete 'module' information, not used by distutils.setup
    del kwargs['module']

    # Write a version.py and add it to the list of files to
    # be installed.
    if write_version:
        f = open('src/version.py', 'w')
        f.write('# version information for %s\n' % kwargs['name'])
        f.write('# autogenerated by kaa.distribution\n\n')
        f.write('from kaa.distribution.version import Version\n')
        f.write('VERSION = Version(\'%s\')\n' % kwargs['version'])
        f.close()

    # add some missing keywords
    if 'author' not in kwargs:
        kwargs['author'] = 'Freevo Development Team'
    if 'author_email' not in kwargs:
        kwargs['author_email'] = 'freevo-devel@lists.sourceforge.net'
    if 'url' not in kwargs:
        kwargs['url'] = 'http://freevo.sourceforge.net/kaa'

    # We use summary and description as keywords that map to distutils
    # description and long_description
    if 'description' in kwargs:
        kwargs['long_description'] = kwargs['description']
        del kwargs['description']
    if 'summary' in kwargs:
        kwargs['description'] = kwargs['summary']
        if 'long_description' not in kwargs:
            kwargs['long_description'] = kwargs['summary']
        del kwargs['summary']

    # add extra commands
    if not 'cmdclass' in kwargs:
        kwargs['cmdclass'] = {}
    kwargs['cmdclass']['build_py'] = build_py

    if len(sys.argv) > 1 and sys.argv[1] == 'bdist_rpm':
        dist = None
        kwargs['name'] = 'python-' + kwargs['name']
        release = "1"
        if '--dist' in sys.argv:
            # TODO: determine this automatically
            idx = sys.argv.index('--dist')
            sys.argv.pop(idx)
            dist = sys.argv.pop(idx)

        if '--release' in sys.argv:
            idx = sys.argv.index('--release')
            sys.argv.pop(idx)
            release = sys.argv.pop(idx)

        if '--snapshot' in sys.argv:
            # If --snapshot is specified on the command line, set the release
            # to contain today's date, for bundling svn snapshots.
            release = "0.%s" % time.strftime("%Y%m%d")
            sys.argv.remove('--snapshot')

        if dist:
            release += "." + dist

        sys.argv.append('--release=%s' % release)

        if 'rpminfo' in kwargs:
            # Grab rpm metadata from setup kwargs and expose as cmdline 
            # parameters to distutils.
            rpminfo = kwargs['rpminfo']
            if dist in rpminfo:
                # dist-specific parameters take precedence
                for key, value in rpminfo[dist].items():
                    rpminfo[key] = value
            for param in ('requires', 'build_requires', 'conflicts', 'obsoletes', 'provides'):
                if param in rpminfo:
                    sys.argv.append("--%s=%s" % (param.replace('_', '-'), rpminfo[param]))


    if 'rpminfo' in kwargs:
        del kwargs['rpminfo']

    # run the distutils.setup function
    return distutils.core.setup(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Rob Shortt <rob@tvcentric.com>
#
# A notifier implementation using Twisted.
#
# Copyright (C) 2008
# Rob Shortt <rob@tvcentric.com>
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""
This is a notifier implementation using Twisted - http://www.twistedmatrix.com/
Twisted is an async framework that has much in common with pynotifier and kaa.

Here are some links of interest to aid development and debuging:

The reactor base class, posixbase, and selectreactor:
http://twistedmatrix.com/trac/browser/trunk/twisted/internet/base.py
http://twistedmatrix.com/trac/browser/trunk/twisted/internet/posixbase.py
http://twistedmatrix.com/trac/browser/trunk/twisted/internet/selectreactor.py

Timers and scheduling:
http://twistedmatrix.com/projects/core/documentation/howto/time.html

Twisted doc index:
http://twistedmatrix.com/projects/core/documentation/howto/index.html
"""

# Python imports
from types import IntType

# Twisted uses zope.interface
from zope.interface import implements

# Twisted imports
from twisted.internet.interfaces import IReadDescriptor, IWriteDescriptor
from twisted.internet import reactor
from twisted.internet import task

# internal packages
import dispatch
import log

IO_READ = 1
IO_WRITE = 2
IO_EXCEPT = 4

__sockobjs = {}
__sockobjs[IO_READ] = {}
__sockobjs[IO_WRITE] = {}
__timers = {}
__timer_id = 0
__dispatch_timer = None


class SocketReadCB:
    """
    An object to implement Twisted's IReadDescriptor.  When there is data 
    available on the socket doRead() will get called.
    """
    implements(IReadDescriptor)

    def __init__(self, socket, method):
        self.socket = socket
        self.method = method

    def doRead(self):
        """
        Call the callback method with the socket as the only argument.  If it
        returns False remove this socket from our notifier.
        """
        if self.method(self.socket) == False:
            socket_remove(self.socket, IO_READ)

    def fileno(self):
        if type(self.socket) is IntType:
            return self.socket
        elif hasattr(self.socket, 'fileno'):
            return self.socket.fileno()

    def logPrefix(self):
        return "notifier"

    def connectionLost(self, reason): 
        # Should we do more?
        log.error("connection lost on socket fd=%s" % self.fileno())


class SocketWriteCB:
    """
    An object to implement Twisted's IWriteDescriptor.  When there is data 
    available on the socket doWrite() will get called.
    """
    implements(IWriteDescriptor)

    def __init__(self, socket, method):
        self.socket = socket
        self.method = method

    def doWrite(self):
        """
        Call the callback method with the socket as the only argument.  If it
        returns False remove this socket from our notifier.
        """
        if self.method(self.socket) == False:
            socket_remove(self.socket, IO_WRITE)

    def fileno(self):
        if type(self.socket) is IntType:
            return self.socket
        elif hasattr(self.socket, 'fileno'):
            return self.socket.fileno()

    def logPrefix(self):
        return "notifier"

    def connectionLost(self, reason): 
        # Should we do more?
        log.error("connection lost on socket fd=%s" % self.fileno())


def socket_add(id, method, condition = IO_READ):
    """
    The first argument specifies a socket, the second argument has to be a
    function that is called whenever there is data ready in the socket.

    Objects that implement Twisted's IRead/WriteDescriptor interfaces get
    passed to the reactor to monitor.
    """
    global __sockobjs

    if condition == IO_READ:
        s = SocketReadCB(id, method)
        reactor.addReader(s)

    elif condition == IO_WRITE:
        s = SocketWriteCB(id, method)
        reactor.addWriter(s)

    else:
        return

    __sockobjs[condition][id] = s


def socket_remove(id, condition=IO_READ):
    """
    Removes the IRead/WriteDescriptor object with this socket from
    the Twisted reactor.
    """
    global __sockobjs
    sockobj = __sockobjs[condition].get(id)

    if sockobj:
        if condition == IO_READ:
            reactor.removeReader(sockobj)
        elif condition == IO_WRITE:
            reactor.removeWriter(sockobj)
        del __sockobjs[condition][id]


def timer_add(interval, method):
    """
    The first argument specifies an interval in milliseconds, the second
    argument a function. This is function is called after interval
    seconds. If it returns true it's called again after interval
    seconds, otherwise it is removed from the scheduler. The third
    (optional) argument is a parameter given to the called
    function. This function returns an unique identifer which can be
    used to remove this timer
    """
    global __timer_id

    try:
        __timer_id += 1
    except OverflowError:
        __timer_id = 0

    t = task.LoopingCall(method)
    t.start(interval/1000.0, now=False)
    __timers[__timer_id] = t 

    return __timer_id


def timer_remove(id):
    """
    Removes the timer identifed by the unique ID from the main loop.
    """
    t = __timers.get(id)
    if t != None:
        t.stop()
        del __timers[id]


def dispatcher_add(method):
    dispatch.dispatcher_add(method)

dispatcher_remove = dispatch.dispatcher_remove


def step(sleep = True, external = True):
    if reactor.running:
        try:
            t = sleep and reactor.running and reactor.timeout()
            reactor.doIteration(t)
            reactor.runUntilCurrent()
        except:
            log.error("problem running reactor - exiting")
            raise SystemExit
        if external:
            dispatch.dispatcher_run()
    else:
        log.info("reactor stopped - exiting")
        raise SystemExit


def loop():
    """
    Instead of calling reactor.run() here we must call step() so we get a
    chance to call dispatch.dispatcher_run().  Otherwise the dispatchers 
    would have to be run in a timer, making something like the 'step' signal
    not getting called every iteration of the main loop like it was intended.

    We could also decide between reactor.run() and step() and if we use step()
    just setup a Timer for the dispatchers at a reasonable rate.  ie:

        global __dispatch_timer
        __dispatch_timer = task.LoopingCall(dispatch.dispatcher_run)
        __dispatch_timer.start(dispatch.MIN_TIMER/1000.0) # 10x / second
        # or
        # __dispatch_timer.start(1.0/30) # 30x / second
    """
    while True:
        try:
            step()
        except:
            log.debug("exiting loop")
            break


def _init():
    reactor.startRunning(installSignalHandlers=True)




# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# main.py - Main loop functions
# -----------------------------------------------------------------------------
# $Id: main.py 3122 2008-02-21 10:21:38Z dmeyer $
#
# -----------------------------------------------------------------------------
# kaa.notifier - Mainloop and callbacks
# Copyright (C) 2005-2008 Dirk Meyer, Jason Tackaberry, et al.
#
# First Version: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#                Jason Tackaberry <tack@urandom.ca>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

# python imports
import sys
import logging
import os
import time
import signal
import threading
import atexit

import nf_wrapper as notifier
from signals import Signal
from timer import OneShotTimer
from popen import proclist as _proclist
from thread import is_mainthread, wakeup, set_as_mainthread, threaded, MAINTHREAD
from thread import killall as kill_jobserver

__all__ = [ 'run', 'stop', 'step', 'select_notifier', 'is_running', 'wakeup',
            'set_as_mainthread', 'is_shutting_down', 'loop' ]

# get logging object
log = logging.getLogger('notifier')

# pid of the process running the notifier loop.  This lets run() know
# if we've just forked and we want to run a new loop.
_running_pid = None
# Set if currently in shutdown() (to prevent reentrancy)
_shutting_down = False

def _step_signal_changed(signal, flag):
    if flag == Signal.SIGNAL_CONNECTED and signal.count() == 1:
        notifier.dispatcher_add(signals["step"].emit)
    elif flag == Signal.SIGNAL_DISCONNECTED and signal.count() == 0:
        notifier.dispatcher_remove(signals["step"].emit)

signals = {
    'exception': Signal(),
    'shutdown': Signal(),
    'step': Signal(changed_cb = _step_signal_changed),
}


def select_notifier(module, **options):
    """
    Initialize the specified notifier.
    """
    if module in ('thread', 'twisted'):
        import nf_thread
        return nf_thread.init(module, **options)
    return notifier.init( module, **options )


def loop(condition, timeout = None):
    """
    Executes the main loop until condition is met.  condition is either a
    callable, or value that is evaluated after each step of the main loop.
    """
    initial_mainloop = False
    if not is_running():
        # no mainloop is running, set this thread as mainloop and
        # set the internal running state.
        initial_mainloop = True
        set_as_mainthread()
        _set_running(True)

    if not callable(condition):
        condition = lambda: condition

    abort = []
    if timeout is not None:
        # timeout handling to stop the mainloop after the given timeout
        # even when the condition is still True.
        timeout = OneShotTimer(lambda: abort.append(True))
        timeout.start(timeout)

    try:
        while condition() and not abort:
            try:
                notifier.step()
            except Exception, e:
                if signals['exception'].emit(*sys.exc_info()) != False:
                    # Either there are no global exception handlers, or none of
                    # them explicitly returned False to abort mainloop 
                    # termination.  So abort the main loop.
                    type, value, tb = sys.exc_info()
                    raise type, value, tb
    finally:
        # make sure we set mainloop status even for Exceptions we did not
        # catch. E.g. in Python 2.5 SystemExit and KeyboardInterrupt do not
        # inherit from Exception.
        if timeout is not None:
            timeout.stop(timeout)
        if initial_mainloop:
            _set_running(False)


def run():
    """
    Notifier main loop function. It will loop until an exception
    is raised or sys.exit is called.
    """
    if is_running():
        raise RuntimeError('Mainthread is already running')

    try:
        # Nested try necessary because python 2.4 doesn't support
        # try/except/finally.
        try:
            loop(True)
        except (KeyboardInterrupt, SystemExit):
            try:
                # This looks stupid, I know that. The problem is that if we have
                # a KeyboardInterrupt, that flag is still valid somewhere inside
                # python. The next system call will fail because of that. Since we
                # don't want a join of threads or similar fail, we use a very short
                # sleep here. In most cases we won't sleep at all because this sleep
                # fails. But after that everything is back to normal.
                time.sleep(0.001)
            except:
                pass
    finally:
        stop()


# Ensure stop() is called from main thread.
@threaded(MAINTHREAD)
def stop():
    """
    Shutdown notifier and kill all background processes.
    """
    global _shutting_down

    if _shutting_down:
        return

    if is_running():
        # notifier loop still running, send system exit
        log.info('Stop notifier loop')
        notifier.shutdown()

    _shutting_down = True

    _proclist.stop_all()
    signals["shutdown"].emit()
    signals["shutdown"].disconnect_all()
    signals["step"].disconnect_all()

    # Kill processes _after_ shutdown emits to give callbacks a chance to
    # close them properly.
    _proclist.kill_all()
    while _proclist.check():
        # wait until all processes are stopped
        step()
    kill_jobserver()
    # Collect any zombies
    try:
        os.waitpid(-1, os.WNOHANG)
    except:
        pass


def step(*args, **kwargs):
    """
    Notifier step function with signal support.
    """
    if not is_mainthread():
        # If step is being called from a thread, wake up the mainthread
        # instead of allowing the thread into notifier.step.
        wakeup()
        # Sleep for epsilon to prevent busy loops.
        time.sleep(0.001)
        return

    try:
        notifier.step(*args, **kwargs)
    except (KeyboardInterrupt, SystemExit):
        raise SystemExit


def is_running():
    """
    Return if the main loop is currently running.
    """
    return _running_pid == os.getpid()


def is_shutting_down():
    """
    Return if the mainloop is currently inside stop()
    """
    return _shutting_down


def _set_running(status):
    """
    Set mainloop running status.
    """
    global _running_pid
    if status:
        _running_pid = os.getpid()
    else:
        _running_pid = None


def _shutdown_check(*args):
    # Helper function to shutdown kaa on system exit
    # The problem is that pytgtk just exits python and
    # does not simply return from the main loop and kaa
    # can't call the shutdown handler. This is not a perfect
    # solution, e.g. with the generic notifier you can do
    # stuff after kaa.main.run() which is not possible with gtk
    if is_running():
        # If the kaa mainthread (i.e. thread the mainloop is running in)
        # is not the program's main thread, then is_mainthread() will be False
        # and we don't need to set running=False since shutdown() will raise a
        # SystemExit and things will exit normally.
        if is_mainthread():
            _set_running(False)
        stop()


# catch SIGTERM and SIGINT if possible for a clean shutdown
if threading.enumerate()[0] == threading.currentThread():
    def signal_handler(*args):
        # use the preferred stop function for the notifier. Most notifier
        # backends only call sys.exit(0). Some, like twisted need specific code.
        notifier.shutdown()
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)
else:
    log.info('kaa imported from thread, disable SIGTERM handler')
    
# check to make sure we really call our shutdown function
atexit.register(_shutdown_check)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# xmlutils.py - some classes helping dealing with xml files
# -----------------------------------------------------------------------------
# $Id: xmlutils.py 2718 2007-06-23 09:21:47Z dmeyer $
#
# -----------------------------------------------------------------------------
# Copyright (C) 2007 Dirk Meyer, Jason Tackaberry
#
# First Edition: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'SaxTreeHandler' ]

# python xml import
try:
    import xml.sax
except ImportError, e:
    # Small hack before the next release. FIXME: this should be
    # removed for future versions again.
    print 'Error: detected files from previous kaa.base version!'
    print 'Please reinstall kaa by deleting all \'build\' subdirs'
    print 'for all kaa source modules and delete the kaa directory'
    print 'in the site-package directory'
    raise e

class SaxTreeHandler(xml.sax.ContentHandler):
    """
    Handler for the SAX parser. The member function 'handle' will
    be called everytime an element given on init is closed. The parameter
    is the tree with this element as root. A node can either have children
    or text content. The SaxTreeHandler is usefull for simple xml files
    found in config files and information like epg data.
    """
    class Node(object):
        """
        A node created by the SaxTreeHandler
        """
        def __init__(self, name, attr):
            self.name = name
            self.attr = attr
            self.children = []
            self.content = ''

        def getattr(self, attr):
            if self.attr.has_key(attr):
                return self.attr.get(attr)
            nodes = [ n for n in self.children if n.name == attr ]
            if len(nodes) == 1:
                return nodes[0]
            return None

        def __repr__(self):
            return '<Node %s>' % self.name

    def __init__(self, *elements):
        """
        Create handler with a list of element names.
        """
        self._elements = elements
        self._nodes = []


    def startElement(self, name, attr):
        """
        SAX callback
        """
        node = self.Node(name, attr)
        if len(self._nodes):
            self._nodes[-1].children.append(node)
        self._nodes.append(node)


    def endElement(self, name):
        """
        SAX callback
        """
        node = self._nodes.pop()
        node.content = node.content.strip()
        if name in self._elements:
            self.handle(node)
        if not self._nodes:
            self.finalize()


    def characters(self, c):
        """
        SAX callback
        """
        if len(self._nodes):
            self._nodes[-1].content += c


    def handle(self, node):
        """
        SaxTreeHandler callback for a complete node.
        """
        pass


    def finalize(self):
        """
        SaxTreeHandler callback at the end of parsing.
        """
        pass


    def parse(self, f):
        parser = xml.sax.make_parser()
        parser.setFeature(xml.sax.handler.feature_external_ges, False)
        parser.setContentHandler(self)
        parser.parse(f)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# inotify/__init__.py - Inotify interface
# -----------------------------------------------------------------------------
# $Id: __init__.py 3069 2008-02-13 18:43:13Z dmeyer $
#
# -----------------------------------------------------------------------------
# Copyright (C) 2006 Jason Tackaberry, Dirk Meyer
#
# First Edition: Jason Tackaberry <tack@urandom.ca>
# Maintainer:    Jason Tackaberry <tack@urandom.ca>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

# python imports
import os
import struct
import logging
import fcntl
import select

# kaa imports
import kaa

try:
    # imports C module
    import _inotify
except ImportError:
    _inotify = None

# get logging object
log = logging.getLogger('inotify')

# TODO: hook in gamin if it is running. See gamin.py

class INotify(object):
    __instance = None  # Singleton

    def __new__(cls):
        if not INotify.__instance:
            INotify.__instance = object.__new__(cls)

        return INotify.__instance


    def __init__(self):
        if hasattr(self, '_fd'):
            # Constructor already called.  (We may be called again implicitly
            # due to singleton behaviour in __new__
            return

        if not _inotify:
            self._fd = -1
            raise SystemError, "INotify support not compiled."

        self.signals = {
            # Master signal: this signal gets emitted on all events.
            "event": kaa.Signal()
        }

        self._watches = {}
        self._watches_by_path = {}
        # We keep track of recently removed watches so we don't get confused
        # if an event callback removes a watch while we're currently
        # processing a batch of events and we receive an event for a watch
        # we just removed.
        self._watches_recently_removed = []
        self._read_buffer = ""
        self._move_state = None  # For MOVED_FROM events
        self._moved_timer = kaa.WeakOneShotTimer(self._emit_last_move)

        self._fd = _inotify.init()

        if self._fd < 0:
            raise SystemError, "INotify support not detected on this system."

        fcntl.fcntl(self._fd, fcntl.F_SETFL, os.O_NONBLOCK)
        self._mon = kaa.WeakIOMonitor(self._handle_data)
        self._mon.register(self._fd)


    def __del__(self):
        if os and self._fd >= 0 and self._mon:
            os.close(self._fd)
            self._mon.unregister()
            self._mon = None


    def watch(self, path, mask = None):
        """
        Adds a watch to the given path.  The default mask is anything that
        causes a change (new file, deleted file, modified file, or attribute
        change on the file).  This function returns a Signal that the caller
        can then connect to.  Any time there is a notification, the signal
        will get emitted.  Callbacks connected to the signal must accept 2
        arguments: notify mask and filename.
        """
        path = os.path.realpath(path)
        if path in self._watches_by_path:
            return self._watches_by_path[path][0]

        if mask == None:
            mask = INotify.WATCH_MASK

        wd = _inotify.add_watch(self._fd, path, mask)
        if wd < 0:
            raise IOError, "Failed to add watch on '%s'" % path

        signal = kaa.Signal()
        self._watches[wd] = [signal, path]
        self._watches_by_path[path] = [signal, wd]
        return signal


    def ignore(self, path):
        """
        Removes a watch on the given path.
        """
        path = os.path.realpath(path)
        if path not in self._watches_by_path:
            return False

        wd = self._watches_by_path[path][1]
        _inotify.rm_watch(self._fd, wd)
        del self._watches[wd]
        del self._watches_by_path[path]
        self._watches_recently_removed.append(wd)
        return True


    def has_watch(self, path):
        """
        Return if the given path is currently watched by the inotify
        object.
        """
        path = os.path.realpath(path)
        return path in self._watches_by_path


    def get_watches(self):
        """
        Returns a list of all paths monitored by the object.
        """
        return self._watches_by_path.keys()


    def _emit_last_move(self):
        """
        Emits the last move event (MOVED_FROM), if it exists.
        """
        if not self._move_state:
            return

        prev_wd, prev_mask, dummy, prev_path = self._move_state
        self._watches[prev_wd][0].emit(prev_mask, prev_path)
        self.signals["event"].emit(prev_mask, prev_path)
        self._move_state = None
        self._moved_timer.stop()


    def _handle_data(self):
        try:
            data = os.read(self._fd, 32768)
        except socket.error, (err, msg):
            if err == errno.EAGAIN:
                # Resource temporarily unavailable -- we are trying to read
                # data on a socket when none is avilable.  This should not
                # happen under normal circumstances, so log an error.
                log.error("INotify data handler called but no data available.")
            return

        self._read_buffer += data

        event_len = struct.calcsize('LLLL')
        while True:
            if len(self._read_buffer) < event_len:
                if self._move_state:
                    # We received a MOVED_FROM event with no matching 
                    # MOVED_TO.  If we don't get a matching MOVED_TO in 0.1
                    # seconds, emit the MOVED_FROM event.
                    self._moved_timer.start(0.1)
                break

            wd, mask, cookie, size = struct.unpack("LLLL", self._read_buffer[0:event_len])
            if size:
                name = self._read_buffer[event_len:event_len+size].rstrip('\0')
            else:
                name = None

            self._read_buffer = self._read_buffer[event_len+size:]
            if wd not in self._watches:
                if wd not in self._watches_recently_removed:
                    # Weird, received an event for an unknown watch; this
                    # shouldn't happen under sane circumstances, so log this as
                    # an error.
                    log.error("INotify received event for unknown watch.")
                continue

            path = self._watches[wd][1]
            if name:
                path = os.path.join(path, name)

            if self._move_state:
                # Last event was a MOVED_FROM. So if this is a MOVED_TO and the
                # cookie matches, emit once specifying both paths. If not, 
                # emit two signals.
                if mask & INotify.MOVED_TO and cookie == self._move_state[2]:
                    # Great, they match. Fire a MOVE signal with both paths.
                    # Use the all three signals (global, from, to).
                    mask |= INotify.MOVED_FROM
                    prev_wd, dummy, dummy, prev_path = self._move_state
                    self._watches[prev_wd][0].emit(mask, prev_path, path)
                    self._watches[wd][0].emit(mask, prev_path, path)
                    self.signals["event"].emit(mask, prev_path, path)
                    self._move_state = None
                    self._moved_timer.stop()
                    continue

                # No match, fire the earlier MOVED_FROM signal now 
                # with no target.
                self._emit_last_move()

            if mask & INotify.MOVED_FROM:
                # This is a MOVED_FROM. Don't emit the signals now, let's wait
                # for a MOVED_TO, which we expect to be next.
                self._move_state = wd, mask, cookie, path
                continue

            self._watches[wd][0].emit(mask, path)
            self.signals["event"].emit(mask, path)

            if mask & INotify.IGNORED:
                # Self got deleted, so remove the watch data.
                del self._watches[wd]
                del self._watches_by_path[path]
                self._watches_recently_removed.append(wd)

        if not self._read_buffer and len(self._watches_recently_removed) and \
           not select.select([self._fd], [], [], 0)[0]:
            # We've processed all pending inotify events.  We can reset the
            # recently removed watches list.
            self._watches_recently_removed = []


if _inotify:
    # Copy constants from _inotify to INotify
    for attr in dir(_inotify):
        if attr[0].isupper():
            setattr(INotify, attr, getattr(_inotify, attr))

    INotify.WATCH_MASK = INotify.MODIFY | INotify.ATTRIB | INotify.DELETE | \
                         INotify.CREATE | INotify.DELETE_SELF | \
                         INotify.UNMOUNT | INotify.MOVE | INotify.MOVE_SELF | \
                         INotify.MOVED_FROM | INotify.MOVED_TO

    INotify.CHANGE     = INotify.MODIFY | INotify.ATTRIB




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Andreas Büsching <crunchy@bitkipper.net>
#
# package initialisation
#
# Copyright (C) 2004, 2005, 2006
#		Andreas Büsching <crunchy@bitkipper.net>
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Simple mainloop that watches sockets and timers."""

from version import *

from select import select

import log

socket_add = None
socket_remove = None

timer_add = None
timer_remove = None

dispatcher_add = None
dispatcher_remove = None

loop = None
step = None

# notifier types
( GENERIC, QT, GTK, WX, TWISTED ) = range( 5 )

# socket conditions
IO_READ = None
IO_WRITE = None
IO_EXCEPT = None

def init( model = GENERIC, **kwargs ):
	global timer_add
	global socket_add
	global dispatcher_add
	global timer_remove
	global socket_remove
	global dispatcher_remove
	global loop, step
	global IO_READ, IO_WRITE, IO_EXCEPT

	if model == GENERIC:
		import nf_generic as nf_impl
	elif model == QT:
		import nf_qt as nf_impl
	elif model == GTK:
		import nf_gtk as nf_impl
	elif model == WX:
		import nf_wx as nf_impl
		log.warn( 'the WX notifier is deprecated and is no longer maintained' )
	elif model == TWISTED:
		import nf_twisted as nf_impl
		log.info("using nf_twisted")
	else:
		raise Exception( 'unknown notifier model' )

	socket_add = nf_impl.socket_add
	socket_remove = nf_impl.socket_remove
	timer_add = nf_impl.timer_add
	timer_remove = nf_impl.timer_remove
	dispatcher_add = nf_impl.dispatcher_add
	dispatcher_remove = nf_impl.dispatcher_remove
	loop = nf_impl.loop
	step = nf_impl.step
	IO_READ = nf_impl.IO_READ
	IO_WRITE = nf_impl.IO_WRITE
	IO_EXCEPT = nf_impl.IO_EXCEPT

	if hasattr( nf_impl, '_options' ) and type( nf_impl._options ) == dict:
		for k, v in kwargs.items():
			if nf_impl._options.has_key( k ):
				nf_impl._options[ k ] = v

	if hasattr( nf_impl, '_init' ):
		nf_impl._init()

class Callback:
	def __init__( self, function, *args ):
		self._function = function
		self._args = args

	def __call__( self, *args ):
		tmp = list( args )
		if self._args:
			tmp.extend( self._args )
		if tmp:
			return self._function( *tmp )
		else:
			return self._function()

	def __cmp__( self, rvalue ):
		if not callable( rvalue ): return -1

		if ( isinstance( rvalue, Callback ) and \
			   self._function == rvalue._function ) or \
			   self._function == rvalue:
			return 0

		return -1

	def __nonzero__( self ):
		return bool( self._function )

	def __hash__( self ):
		return self._function.__hash__()



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# callback.py - Callback classes for the notifier
# -----------------------------------------------------------------------------
# $Id: callback.py 3073 2008-02-13 19:08:08Z dmeyer $
#
# -----------------------------------------------------------------------------
# kaa.notifier - Mainloop and callbacks
# Copyright (C) 2005-2008 Dirk Meyer, Jason Tackaberry, et al.
#
# First Version: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#                Jason Tackaberry <tack@urandom.ca>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

__all__ = [ 'Callback', 'WeakCallback' ]

# Python imports
import _weakref
import types
import sys
import logging
import atexit

# get logging object
log = logging.getLogger('notifier')

# Variable that is set to True (via atexit callback) when python interpreter
# is in the process of shutting down.  If we're interested if the interpreter
# is shutting down, we don't want to test that this variable is True, but
# rather that it is not False, because as it is prefixed with an underscore,
# the interpreter might already have deleted this variable in which case it
# is None.
_python_shutting_down = False


def weakref_data(data, destroy_cb = None):
    if type(data) in (str, int, long, types.NoneType, types.FunctionType):
        # Naive optimization for common immutable cases.
        return data
    elif type(data) == types.MethodType:
        cb = WeakCallback(data)
        if destroy_cb:
            cb.set_weakref_destroyed_cb(destroy_cb)
            cb.set_ignore_caller_args()
        return cb
    elif type(data) in (list, tuple):
        d = []
        for item in data:
            d.append(weakref_data(item, destroy_cb))
        if type(data) == tuple:
            d = tuple(d)
        return d
    elif type(data) == dict:
        d = {}
        for key, val in data.items():
            d[weakref_data(key)] = weakref_data(val, destroy_cb)
        return d
    else:
        try:
            if destroy_cb:
                return _weakref.ref(data, destroy_cb)
            return _weakref.ref(data)
        except TypeError:
            pass

    return data

def unweakref_data(data):
    if type(data) in (str, int, long, types.NoneType):
        # Naive optimization for common immutable cases.
        return data
    elif type(data) == _weakref.ReferenceType:
        return data()
    elif type(data) == WeakCallback:
        return data._get_callback()
    elif type(data) in (list, tuple):
        d = []
        for item in data:
            d.append(unweakref_data(item))
        if type(data) == tuple:
            d = tuple(d)
        return d
    elif type(data) == dict:
        d = {}
        for key, val in data.items():
            d[unweakref_data(key)] = unweakref_data(val)
        return d
    else:
        return data



class Callback(object):
    """
    Wrapper for functions calls with arguments inside the notifier. The
    function passed to this objects get the parameter passed to this object
    and after that the args and kwargs defined in the init function.
    """
    def __init__(self, callback, *args, **kwargs):
        assert(callable(callback))
        self._callback = callback
        self._args = args
        self._kwargs = kwargs
        self._ignore_caller_args = False
        self._user_args_first = False


    def set_ignore_caller_args(self, flag = True):
        self._ignore_caller_args = flag


    def set_user_args_first(self, flag = True):
        self._user_args_first = flag


    def get_user_args(self):
        return self._args, self._kwargs


    def _get_callback(self):
        return self._callback


    def _merge_args(self, args, kwargs):
        user_args, user_kwargs = self.get_user_args()
        if self._ignore_caller_args:
            cb_args, cb_kwargs = user_args, user_kwargs
        else:
            if self._user_args_first:
                cb_args, cb_kwargs = user_args + args, kwargs
                cb_kwargs.update(user_kwargs)
            else:
                cb_args, cb_kwargs = args + user_args, user_kwargs
                cb_kwargs.update(kwargs)

        return cb_args, cb_kwargs


    def __call__(self, *args, **kwargs):
        """
        Call the callback function.
        """
        cb = self._get_callback()
        cb_args, cb_kwargs = self._merge_args(args, kwargs)
        if not cb:
            # Is it wise to fail so gracefully here?
            return

        self._entered = True
        result = cb(*cb_args, **cb_kwargs)
        self._entered = False
        return result


    def __repr__(self):
        """
        Convert to string for debug.
        """
        return '<%s for %s>' % (self.__class__.__name__, self._callback)


    def __deepcopy__(self, memo):
        """
        Disable deepcopying because deepcopy can't deal with callables.
        """
        return None

    def __eq__(self, func):
        """
        Compares the given function with the callback function we're wrapping.
        """
        return id(self) == id(func) or self._get_callback() == func


class WeakCallback(Callback):

    def __init__(self, callback, *args, **kwargs):
        super(WeakCallback, self).__init__(callback, *args, **kwargs)
        if type(callback) == types.MethodType:
            # For methods
            self._instance = _weakref.ref(callback.im_self, self._weakref_destroyed)
            self._callback = callback.im_func.func_name
        else:
            self._instance = None
            # Don't weakref lambdas.
            if not hasattr(callback, 'func_name') or callback.func_name != '<lambda>':
                self._callback = _weakref.ref(callback, self._weakref_destroyed)

        self._args = weakref_data(args, self._weakref_destroyed)
        self._kwargs = weakref_data(kwargs, self._weakref_destroyed)
        self._weakref_destroyed_user_cb = None


    def __repr__(self):
        if self._instance and self._instance():
            name = "method %s of %s" % (self._callback, self._instance())
        else:
            name = self._callback
        return '<%s for %s>' % (self.__class__.__name__, name)

    def _get_callback(self):
        if self._instance:
            if self._instance() != None:
                return getattr(self._instance(), self._callback)
        elif isinstance(self._callback, _weakref.ReferenceType):
            return self._callback()
        else:
            return self._callback

    def get_user_args(self):
        return unweakref_data(self._args), unweakref_data(self._kwargs)

    def __call__(self, *args, **kwargs):
        if _python_shutting_down != False:
            # Shutdown
            return False

        return super(WeakCallback, self).__call__(*args, **kwargs)


    def set_weakref_destroyed_cb(self, callback):
        self._weakref_destroyed_user_cb = callback


    def _weakref_destroyed(self, object):
        if _python_shutting_down != False:
            # Shutdown
            return
        try:
            if self._weakref_destroyed_user_cb:
                return self._weakref_destroyed_user_cb(object)
        except:
            log.exception("Exception raised during weakref destroyed callback")



def _shutdown_weakref_destroyed():
    global _python_shutting_down
    _python_shutting_down = True

atexit.register(_shutdown_weakref_destroyed)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Andreas Büsching <crunchy@bitkipper.net>
#
# notifier wrapper for GTK+ 2.x
#
# Copyright (C) 2004, 2005, 2006
#		Andreas Büsching <crunchy@bitkipper.net>
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Simple mainloop that watches sockets and timers."""

import gobject

import dispatch
import log

IO_READ = gobject.IO_IN
IO_WRITE = gobject.IO_OUT
IO_EXCEPT = gobject.IO_ERR

_options = {
	'x11' : True,
}

# map of Sockets/Methods -> GTK source IDs
_gtk_socketIDs = {}
_gtk_socketIDs[ IO_READ ] = {}
_gtk_socketIDs[ IO_WRITE ] = {}

def socket_add( socket, method, condition = IO_READ ):
	"""The first argument specifies a socket, the second argument has to be a
	function that is called whenever there is data ready in the socket."""
	global _gtk_socketIDs
	source = gobject.io_add_watch( socket, condition,
								   _socket_callback, method )
	_gtk_socketIDs[ condition ][ socket ] = source

def _socket_callback( source, condition, method ):
	"""This is an internal callback function, that maps the GTK source IDs
	to the socket objects that are used by pynotifier as an identifier
	"""
	global _gtk_socketIDs
	if _gtk_socketIDs[ condition ].has_key( source ):
		ret = method( source )
		if not ret:
			socket_remove( source, condition )
	return ret

	log.info( "socket '%s' not found" % source )
	return False

def socket_remove( socket, condition = IO_READ ):
	"""Removes the given socket from scheduler."""
	global _gtk_socketIDs
	if _gtk_socketIDs[ condition ].has_key( socket ):
		gobject.source_remove( _gtk_socketIDs[ condition ][ socket ] )
		del _gtk_socketIDs[ condition ][ socket ]
	else:
		log.info( "socket '%s' not found" % socket )

def timer_add( interval, method ):
	"""The first argument specifies an interval in milliseconds, the
	second argument a function. This is function is called after
	interval seconds. If it returns true it's called again after
	interval seconds, otherwise it is removed from the scheduler. The
	third (optional) argument is a parameter given to the called
	function."""
	return gobject.timeout_add( interval, method )

def timer_remove( id ):
	"""Removes the timer specified by id from the scheduler."""
	gobject.source_remove( id )

dispatcher_add = dispatch.dispatcher_add
dispatcher_remove = dispatch.dispatcher_remove

_mainloop = None
_step = None

def step( sleep = True, external = True ):
	global _step
	_step( sleep )
	if external:
		dispatch.dispatcher_run()

def loop():
	"""Execute main loop forever."""
	while 1:
		step()

def _init():
	global _step, _mainloop

	if _options[ 'x11' ] == True:
		import gtk
		_step = gtk.main_iteration_do
	else:
		_mainloop = gobject.main_context_default()
		_step = _mainloop.iteration

	gobject.threads_init()



import kaa

@kaa.coroutine()
def new_client(client):
    ip, port = client.address
    print 'New connection from %s:%s' % (ip, port)
    client.write('Hello %s, connecting from port %d\n' % (ip, port))

    remote = kaa.Socket()
    yield remote.connect('www.freevo.org:80')
    remote.write('GET / HTTP/1.0\n\n')
    while remote.connected:
        data = yield remote.read()
        client.write(data)
    client.write('\n\nBye!\n')
    client.close()

server = kaa.Socket()
server.signals['new-client'].connect(new_client)
server.listen(8080)
print "Connect to localhost:8080"
kaa.main.run()



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# nf_wrapper.py - Wrapper to notifier calls to delay the real import
# -----------------------------------------------------------------------------
# $Id: nf_wrapper.py 3122 2008-02-21 10:21:38Z dmeyer $
#
# -----------------------------------------------------------------------------
# kaa.notifier - Mainloop and callbacks
# Copyright (C) 2006-2008 Dirk Meyer, Jason Tackaberry, et al.
#
# First Version: Dirk Meyer <dmeyer@tzi.de>
# Maintainer:    Dirk Meyer <dmeyer@tzi.de>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

# Python imports
import logging
import sys
import atexit

# notifier import
from callback import Callback, WeakCallback
from signals import Signal

# get logging object
log = logging.getLogger('notifier')

# Variable that is set to True (via atexit callback) when python interpreter
# is in the process of shutting down.  If we're interested if the interpreter
# is shutting down, we don't want to test that this variable is True, but
# rather that it is not False, because as it is prefixed with an underscore,
# the interpreter might already have deleted this variable in which case it
# is None.
_python_shutting_down = False


class NotifierCallback(Callback):

    def __init__(self, callback, *args, **kwargs):
        super(NotifierCallback, self).__init__(callback, *args, **kwargs)
        self._id = None

        self.signals = {
            "exception": Signal(),
            "unregistered": Signal()
        }


    def active(self):
        # callback is active if id is not None and python is not shutting down
        # if python is in shutdown, notifier unregister could crash
        return self._id != None and _python_shutting_down == False


    def unregister(self):
        # Unregister callback with notifier.  Must be implemented by subclasses.
        self.signals["unregistered"].emit()
        self._id = None


    def __call__(self, *args, **kwargs):
        if not self._get_callback():
            if self.active():
                self.unregister()
            return False

        # If there are exception handlers for this notifier callback, we
        # catch the exception and pass it to the handler, giving it the
        # opportunity to abort the unregistering.  If no handlers are
        # attached and an exception is raised, it will be propagated up to
        # our caller.
        if self.signals["exception"].count() > 0:
            try:
                ret = super(NotifierCallback, self).__call__(*args, **kwargs)
            except:
                # If any of the exception handlers return True, then the
                # object is not unregistered from the Notifier.  Otherwise
                # ret = False and it will unregister.
                ret = self.signals["exception"].emit(*sys.exc_info())
        else:
            ret = super(NotifierCallback, self).__call__(*args, **kwargs)
        # If Notifier callbacks return False, they get unregistered.
        if ret == False:
            self.unregister()
            return False
        return True


class WeakNotifierCallback(WeakCallback, NotifierCallback):

    def _weakref_destroyed(self, object):
        if _python_shutting_down == False:
            super(WeakNotifierCallback, self)._weakref_destroyed(object)
            self.unregister()


class _Wrapper(object):
    def __init__(self, name):
        self.name = name

    def __call__(self, *args, **kwargs):
        init()
        return globals()[self.name](*args, **kwargs)

dispatcher_add = _Wrapper('dispatcher_add')
dispatcher_remove = _Wrapper('dispatcher_remove')
step = _Wrapper('step')
timer_remove = _Wrapper('timer_remove')
timer_add = _Wrapper('timer_add')
socket_remove = _Wrapper('socket_remove')
socket_add = _Wrapper('socket_add')

def shutdown():
    # prefered way to shut down the system
    sys.exit(0)

# socket wrapper

nf_conditions = []
def _socket_add(id, method, condition = 0):
    return nf_socket_add(id, method, nf_conditions[condition])


def _socket_remove(id, condition = 0):
    return nf_socket_remove(id, nf_conditions[condition])


def init( module = None, force_internal=False, **options ):
    global timer_add
    global socket_add
    global dispatcher_add
    global timer_remove
    global socket_remove
    global dispatcher_remove
    global step
    global nf_socket_remove
    global nf_socket_add
    global nf_conditions
    global shutdown

    if not isinstance(timer_add, _Wrapper):
        raise RuntimeError('notifier already initialized')

    if not 'recursive_depth' in options:
        # default value of 2 is not enough when using async yield stuff
        options['recursive_depth'] = 5

    try:
        if force_internal:
            # pynotifier is not allowed
            raise ImportError()
        import notifier
        if notifier.loop:
            raise RuntimeError('pynotifier already initialized')
    except ImportError:
        # use our own copy of pynotifier
        import pynotifier as notifier

    # find a good main notifier implementation
    if not module:
        module = 'generic'
        if sys.modules.has_key('gtk'):
            # The gtk module is loaded, this means that we will hook
            # ourself into the gtk notifier
            module = 'gtk'

    if not module in ('generic', 'gtk', 'twisted_experimental'):
        raise AttributeError('unsupported notifier %s' % module)

    if module == 'twisted_experimental':
        module = 'twisted'

    # use the selected module
    notifier.init(getattr(notifier, module.upper()), **options)

    # delete basic notifier handler
    log = logging.getLogger('notifier')
    for l in log.handlers:
        log.removeHandler(l)

    timer_remove = notifier.timer_remove
    timer_add = notifier.timer_add

    nf_socket_remove = notifier.socket_remove
    nf_socket_add = notifier.socket_add
    nf_conditions = [ notifier.IO_READ, notifier.IO_WRITE ]
    socket_remove = _socket_remove
    socket_add = _socket_add

    dispatcher_add = notifier.dispatcher_add
    dispatcher_remove = notifier.dispatcher_remove

    step = notifier.step

    if module == 'twisted':
        # special stop handling for twisted
        from twisted.internet import reactor
        shutdown = reactor.stop


def _shutdown_weakref_destroyed():
    global _python_shutting_down
    _python_shutting_down = True

atexit.register(_shutdown_weakref_destroyed)



# -*- coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# logger.py - Updates to the Python logging module
# -----------------------------------------------------------------------------
# $Id: logger.py 2646 2007-04-18 23:52:51Z tack $
#
# This module 'fixes' the Python logging module to accept fixed string and
# unicode arguments. It will also make sure that there is a logging handler
# defined when needed.
#
# -----------------------------------------------------------------------------
# Copyright (C) 2006 Dirk Meyer, Jason Tackaberry
#
# First Edition: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

# Python imports
import logging

# baa.base imports
from strutils import unicode_to_str


def create_logger(level = logging.WARNING):
    """
    Create a simple logging object for applicatins that don't want
    to create a logging handler on their own. You should always have
    a logging object.
    """
    log = logging.getLogger()
    # delete current handler
    for l in log.handlers:
        log.removeHandler(l)

    # Create a simple logger object
    if len(logging.getLogger().handlers) > 0:
        # there is already a logger, skipping
        return

    formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(module)s(%(lineno)s): %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    log.addHandler(handler)


_makeRecord = logging.Logger.makeRecord

def make_record(self, name, level, fn, lno, msg, args, *_args, **_kwargs):
    """
    A special makeRecord class for the logger to convert msg and args into
    strings using the correct encoding if they are unicode strings. This
    function also makes sure we have at least a basic handler.
    """
    if len(self.root.handlers) == 0:
        # create handler, we don't have one
        create_logger()

    # convert message to string
    msg = unicode_to_str(msg)
    # convert args to string
    args = tuple([ unicode_to_str(x) for x in args ])
    # call original function
    return _makeRecord(self, name, level, fn, lno, msg, args, *_args, **_kwargs)

# override makeRecord of a logger by our new function that can handle
# unicode correctly and that will take care of a basic logger.
logging.Logger.makeRecord = make_record



# -* -coding: iso-8859-1 -*-
# -----------------------------------------------------------------------------
# tls.py - TLS support for kaa.notifier based on tlslite
# -----------------------------------------------------------------------------
# $Id: tls.py 3200 2008-03-19 15:11:32Z dmeyer $
#
# This module wraps TLS for client and server based on tlslite. See
# http://trevp.net/tlslite/docs/public/tlslite.TLSConnection.TLSConnection-class.html
# for more information about optional paramater.
#
# -----------------------------------------------------------------------------
# Copyright (C) 2008 Dirk Meyer
#
# First Edition: Dirk Meyer <dischi@freevo.org>
# Maintainer:    Dirk Meyer <dischi@freevo.org>
#
# Please see the file AUTHORS for a complete list of authors.
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version
# 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA
#
# -----------------------------------------------------------------------------

# python imports
import sys
import logging

# import tlslite API to the namespace of this module
from tlslite.api import *

# import tlslite.api to overwrite TLSConnection
import tlslite.api

# kaa imports
import kaa

# get logging object
log = logging.getLogger('tls')


class TLSConnection(tlslite.api.TLSConnection):
    """
    This class wraps a socket and provides TLS handshaking and data transfer.
    It enhances the tlslite version of the class with the same name with
    kaa.notifier support.
    """
    @kaa.coroutine()
    def _iterate_handshake(self, handshake):
        """
        Iterate through the TLS handshake for asynchronous calls using
        kaa.notifier IOMonitor and InProgressCallback.
        """
        try:
            while True:
                n = handshake.next()
                cb = kaa.InProgressCallback()
                disp = kaa.notifier.IOMonitor(cb)
                if n == 0:
                    disp.register(self.sock.fileno(), kaa.notifier.IO_READ)
                if n == 1:
                    disp.register(self.sock.fileno(), kaa.notifier.IO_WRITE)
                yield cb
                disp.unregister()
        except StopIteration:
            pass

    def handshakeClientCert(self, certChain=None, privateKey=None, session=None,
                            settings=None, checker=None):
        """
        Perform a certificate-based handshake in the role of client.
        """
        handshake = tlslite.api.TLSConnection.handshakeClientCert(
            self, certChain=certChain, privateKey=privateKey, session=session,
            settings=settings, checker=checker, async=True)
        return self._iterate_handshake(handshake)

    def handshakeServer(self, sharedKeyDB=None, verifierDB=None, certChain=None,
                        privateKey=None, reqCert=None, sessionCache=None,
                        settings=None, checker=None):
        """
        Start a server handshake operation on the TLS connection.
        """
        handshake = tlslite.api.TLSConnection.handshakeServerAsync(
            self, sharedKeyDB, verifierDB, certChain, privateKey, reqCert,
            sessionCache, settings, checker)
        return self._iterate_handshake(handshake)

    def fileno(self):
        """
        Return socket descriptor. This makes this class feel like a normal
        socket to the IOMonitor.
        """
        return self.sock.fileno()

    def close(self):
        """
        Close the socket.
        """
        if not self.closed:
            # force socket close or this will block
            # on kaa shutdown.
            self.sock.close()
        return tlslite.api.TLSConnection.close(self)


class TLSSocket(kaa.Socket):
    """
    Special version of kaa.Socket with TLS support. On creation the
    connection is NOT encrypted, starttls_client and starttls_server
    must be called to encrypt the connection.
    """
    def __init__(self):
        kaa.Socket.__init__(self)
        self.signals['tls'] = kaa.Signal()
        self._handshake = False

    def _accept(self):
        """
        Accept a new connection and return a new Socket object.
        """
        sock, addr = self._socket.accept()
        client_socket = TLSSocket()
        client_socket.wrap(sock, addr)
        self.signals['new-client'].emit(client_socket)

    def _update_read_monitor(self, signal = None, change = None):
        # FIXME: This function is broken in TLSSocket for two reasons:
        # 1. auto-reconnect while doing a tls handshake is wrong
        #    This could be fixed using self._handshake
        # 2. Passing self._socket to register does not work,
        #    self._socket.fileno() is needed. Always using fileno()
        #    does not work for some strange reason.
        return

    def wrap(self, sock, addr = None):
        """
        Wraps an existing low-level socket object.  addr specifies the address
        corresponding to the socket.
        """
        super(TLSSocket, self).wrap(sock, addr)
        # since _update_read_monitor is deactivated we need to always register
        # the rmon to the notifier.
        if not self._rmon.active():
            self._rmon.register(self._socket.fileno(), kaa.IO_READ)

    def write(self, data):
        """
        Write data to the socket. The data will be delayed while the socket
        is doing the TLS handshake.
        """
        if self._handshake:
            # do not send data while doing a handshake
            return self._write_buffer.append(data)
        return super(TLSSocket, self).write(data)

    def _handle_read(self):
        """
        Callback for new data on the socket.
        """
        try:
            return super(TLSSocket, self)._handle_read()
        except TLSAbruptCloseError, e:
            log.error('TLSAbruptCloseError')
            self._read_signal.emit(None)
            self._readline_signal.emit(None)
            return self.close(immediate=True, expected=False)

    @kaa.coroutine()
    def starttls_client(self, session=None, key=None, **kwargs):
        """
        Start a certificate-based handshake in the role of a TLS client.
        Note: this function DOES NOT check the server key based on the
        key chain. Provide a checker callback to be called for verification.
        http://trevp.net/tlslite/docs/public/tlslite.Checker.Checker-class.html
        Every callable object can be used as checker.
        """
        try:
            if key:
                kwargs['privateKey'] = key.key
                kwargs['certChain'] = key.chain
            self._handshake = True
            if session is None:
                session = Session()
            c = TLSConnection(self._socket)
            self._rmon.unregister()
            yield c.handshakeClientCert(session=session, **kwargs)
            self._socket = c
            self.signals['tls'].emit()
            self._rmon.register(self._socket.fileno(), kaa.IO_READ)
            self._handshake = False
        except:
            self._handshake = False
            type, value, tb = sys.exc_info()
            raise type, value, tb

    @kaa.coroutine()
    def starttls_server(self, key, **kwargs):
        """
        Start a certificate-based handshake in the role of a TLS server.
        Note: this function DOES NOT check the client key if requested,
        provide a checker callback to be called for verification.
        http://trevp.net/tlslite/docs/public/tlslite.Checker.Checker-class.html
        Every callable object can be used as checker.
        """
        try:
            self._handshake = True
            c = TLSConnection(self._socket)
            self._rmon.unregister()
            yield c.handshakeServer(privateKey=key.key, certChain=key.chain, **kwargs)
            self._socket = c
            self.signals['tls'].emit()
            self._rmon.register(self._socket.fileno(), kaa.IO_READ)
            self._handshake = False
        except:
            self._handshake = False
            type, value, tb = sys.exc_info()
            raise type, value, tb


class TLSKey(object):
    """
    Class to hold the public (and private) key together with the certification chain.
    This class can be used with TLSSocket as key.
    """
    def __init__(self, filename, private, *certs):
        self.key = parsePEMKey(open(filename).read(), private=private)
        chain = []
        for cert in (filename, ) + certs:
            x509 = X509()
            x509.parse(open(cert).read())
            chain.append(x509)
        self.chain = X509CertChain(chain)



# -*- coding: iso-8859-1 -*-

from kaa.config import Var, Group, Dict, List, Config, VarProxy


def config_change_cb(name, oldval, newval):
    print "******* CFG ITEM CHANGE", name, oldval, newval

config = Config(desc='basic config group', schema=[
    Var(name='foo', desc='some text', default=5),
    Var(name='bar', default=u'bar',
        desc='more text\ndescription has two lines'),

    # group defined inside the basic schema
    Group(name='inline', desc='this is a subgroup', schema=[
    Var(name='x', desc='desc_x', default=7 ),
    Var(name='y', type=range(0,5), desc='desc_y', default=3 ) ])
    ])


# create extra group and add it to the schema
subgroup = Group(desc='this is a subgroup', schema=[
    Var(name='x', desc=u'desc_x with non ascii ü', default=7 ),
    # the next variable allows numbers from 0-4
    Var(name='y', type=range(0,5), desc='desc_y', default=3 ) ])
config.add_variable('subgroup', subgroup)

# create a group again deeper in the tree
subsubgroup = Group(desc='desrc of subsubgroup', schema=[
    Var(name='a', desc='desc a', default=3 ) ])
subgroup.add_variable('z', subsubgroup)

# create a list of a group
l = List(desc='desrc of list subsubgroup', schema=Group([
    Var(name='a', type=int, desc='desc a', default=3 ),
    # z is again a group
    Group(name='z', desc='this is a subgroup', schema=[
    Var(name='x', desc='desc_x', default=7 ),
    Var(name='y', type=range(0,5), desc='desc_y', default=3 ) ]) ]))
subgroup.add_variable('list', l)

# create a dict of strings
epg = Dict(desc='desrc of dict epg', schema=Var(default=''))
subgroup.add_variable('epg', epg)

# store the schema up to this point, we will need it later
part_config = config.copy()

# create extra group and add it to the schema
subgroup = Group(desc='this is a subgroup', schema=[
    Var(name='x', desc='desc_x', default=7 ) ])
config.add_variable('some_group', subgroup)

# OK, let's play with the config

config.subgroup.list.add_monitor(config_change_cb)
config.add_monitor(config_change_cb)
print '** Test 1: change config.subgroup.list and create some errors **'
print config.subgroup
print config.subgroup.list
print config.subgroup.list[0]
print config.subgroup.list[0].a
config.subgroup.list[0].a = 6
print config.subgroup.list[0].a

# This crashes because there is no .a
try:
    config.subgroup.list[1].z.a = 7
except Exception, e:
    print e

# This crashes because the index is no int
try:
    config.subgroup.list['foo'].z.x = 7
except Exception, e:
    print e

print config.subgroup.list[1].z.x
config.subgroup.list[1].z.x = 8
print config.subgroup.list[1].z.x

print
print '** Test 2: play with the dict **'

epg['foo'] = 'bar'
epg['x']   = u'non-ascii: ö'
epg[u'also-non-ascii ä'] = u'non-ascii: ö'
epg['this.has.a.dot'] = 'something'
epg['this.has.a.='] = 'something'
epg['the other way around'] = 'something=bar'

print epg['foo']
print
for key, value in epg.items():
    print key, value
print
for key in epg:
    print key
for var in epg.values():
    print var
print
print '** Test 3: play some other variables **'

print config.foo

config.foo = 10
print config.foo
# This crashes because hello is no int
try:
    config.foo = 'hello'
except Exception, e:
    print e
print config.subgroup.x
config.subgroup.x = 10
print config.subgroup.x

print 'y', config.subgroup.y
# This crashes because 8 is not in range
try:
    config.subgroup.y = 8
except AttributeError, e:
    print e
print 'y', config.subgroup.y
config.subgroup.y = 2
print 'y', config.subgroup.y

try:
    config.not_there
except Exception, e:
    print e

config.some_group.x = 1

print
print '** Test 4: save and reload **'

print 'config.subgroup.z.a is', config.subgroup.z.a
print 'set to 5'
config.subgroup.z.a = 5
print 'config.subgroup.z.a is', config.subgroup.z.a

print 'save config to filename config.test'
config.save('config.test')
print 'change config object'
config.subgroup.z.a = 6
print 'config.subgroup.z.a is', config.subgroup.z.a
print 'read config file into new object'
if not config.load('config.test'):
    print 'load error, bad lines saved, not expected'
print 'config.subgroup.z.a is', config.subgroup.z.a
print 'load again in incomplete schema'
if not part_config.load('config.test'):
    print 'load error, bad lines saved as expected'
print isinstance(config.subgroup.z.a, str)
part_config.save('config.test2')
